# -*- coding: utf-8 -*-
"""Test global functions in mincepy"""
import mincepy
from mincepy import testing


def test_archive_uri_deprecated(monkeypatch):
    """Test the old, deprecated, version"""
    monkeypatch.delenv(mincepy.ENV_ARCHIVE_URI, raising=False)
    assert mincepy.archive_uri() == mincepy.DEFAULT_ARCHIVE_URI
    monkeypatch.setenv(mincepy.ENV_ARCHIVE_URI, 'mongodb://example.com')
    assert mincepy.archive_uri() == 'mongodb://example.com'


def test_default_archive_uri(monkeypatch):
    monkeypatch.delenv(mincepy.ENV_ARCHIVE_URI, raising=False)
    assert mincepy.default_archive_uri() == mincepy.DEFAULT_ARCHIVE_URI
    monkeypatch.setenv(mincepy.ENV_ARCHIVE_URI, 'mongodb://example.com')
    assert mincepy.default_archive_uri() == 'mongodb://example.com'


def test_save_load():
    car = testing.Car()
    car_id = mincepy.save(car)
    del car
    # Now try loading
    mincepy.load(car_id)


def test_find():
    car = testing.Car()
    car_id = car.save()
    assert list(mincepy.find(obj_id=car_id))[0] is car


def test_delete():
    car = testing.Car()
    car_id = car.save()
    mincepy.delete(car)
    assert not list(mincepy.find(obj_id=car_id))



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod
import datetime
from typing import Type, List
import uuid

try:  # Python3
    from hashlib import blake2b
except ImportError:  # Python < 3.6
    from pyblake2 import blake2b

import mincepy  # pylint: disable=unused-import
from . import depositors
from . import expr
from . import fields
from . import saving
from . import tracking

__all__ = 'Savable', 'Comparable', 'Object', 'SavableObject', 'PRIMITIVE_TYPES'

# The primitives that all archive types must support
PRIMITIVE_TYPES = (bool, int, float, str, dict, list, type(None), bytes, uuid.UUID,
                   datetime.datetime)


def is_primitive(obj):
    return isinstance(obj, PRIMITIVE_TYPES)


class Savable(fields.WithFields, expr.FilterLike):
    """Interface for an object that can save an load its instance state"""
    TYPE_ID = None
    LATEST_MIGRATION = None  # type: mincepy.ObjectMigration

    def __init__(self, *args, **kwargs):
        assert self.TYPE_ID is not None, 'Must set the TYPE_ID for an object to be savable'
        super().__init__(*args, **kwargs)

    @classmethod
    def __expr__(cls):
        """This method gives savables the ability to be used as an expression"""
        return expr.Comparison('type_id', expr.Eq(cls.TYPE_ID))

    @classmethod
    def __query_expr__(cls) -> dict:
        """This method gives savables the ability to be used in query filter expressions"""
        return cls.__expr__().__query_expr__()

    def save_instance_state(self, saver: depositors.Saver):  # pylint: disable=unused-argument
        """Save the instance state of an object, should return a saved instance"""
        return saving.save_instance_state(self)

    def load_instance_state(self, saved_state, loader: depositors.Loader):  # pylint: disable=unused-argument
        """Take the given object and load the instance state into it"""
        saving.load_instance_state(self, saved_state)


class Comparable(metaclass=ABCMeta):
    """Interface for an object that can be compared and hashed"""

    @abstractmethod
    def __eq__(self, other) -> bool:
        """Determine if two objects are equal"""

    @abstractmethod
    def yield_hashables(self, hasher):
        """Produce a hash representing the value"""


class Object(Comparable, metaclass=ABCMeta):
    """A simple object that is comparable"""


class SavableObject(Object, Savable, metaclass=ABCMeta):
    """A class that is both savable and comparable"""

    _historian = None

    @classmethod
    def init_field(cls, field: fields.Field, attr_name: str):
        super().init_field(field, attr_name)
        field.set_query_context(expr.Comparison('type_id', expr.Eq(cls.TYPE_ID)))
        field.path_prefix = 'state'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        tracking.obj_created(self)

    def __eq__(self, other) -> bool:
        """Determine if two objects are equal"""
        if not isinstance(other, type(self)):
            return False

        return saving.save_instance_state(self) == saving.save_instance_state(other)

    def yield_hashables(self, hasher):
        """Produce a hash representing the object"""
        yield from hasher.yield_hashables(saving.save_instance_state(self))


class Equator:

    def __init__(self, equators=tuple()):
        self._equators = list(equators)

        def do_hash(*args):
            hasher = blake2b(digest_size=32)
            for arg in args:
                hasher.update(arg)

            return hasher.hexdigest()

        self._hasher = do_hash

    def add_equator(self, equator):
        self._equators.append(equator)

    def remove_equator(self, equator):
        self._equators.reverse()
        try:
            self._equators.remove(equator)
        except ValueError as exc:
            raise ValueError("Unknown equator '{}'".format(equator)) from exc
        finally:
            self._equators.reverse()

    def get_equator(self, obj):
        # Iterate in reversed order i.e. the latest added should be used preferentially
        for equator in reversed(self._equators):
            if isinstance(obj, equator.TYPE):
                return equator
        raise TypeError("Don't know how to compare '{}' types, no type equator set".format(
            type(obj)))

    def yield_hashables(self, obj):
        try:
            equator = self.get_equator(obj)
        except TypeError:
            # Try the objects's method
            try:
                yield from obj.yield_hashables(self)
            except AttributeError:
                raise TypeError("No helper registered and no yield_hashables method on '{}'".format(
                    type(obj))) from None
        else:
            yield from equator.yield_hashables(obj, self)

    def hash(self, obj):
        return self._hasher(*self.yield_hashables(obj))

    def eq(self, obj1, obj2) -> bool:  # pylint: disable=invalid-name
        if not type(obj1) == type(obj2):  # pylint: disable=unidiomatic-typecheck
            return False

        try:
            equator = self.get_equator(obj1)
        except TypeError:
            # Fallback to python eq
            return obj1 == obj2
        else:
            return equator.eq(obj1, obj2)

    def float_to_str(self, value, sig=14):  # pylint: disable=no-self-use
        """
        Convert float to text string for computing hash.
        Preserve up to N significant number given by sig.

        :param value: the float value to convert
        :param sig: choose how many digits after the comma should be output
        """
        fmt = u'{{:.{}g}}'.format(sig)
        return fmt.format(value)


def is_savable_type(obj_type: Type) -> bool:
    return issubclass(obj_type, SavableObject) and obj_type.TYPE_ID is not None


def savable_mro(obj_type: Type[SavableObject]) -> List[Type[SavableObject]]:
    """Given a SavableObject type this will give the mro of the savable types in the hierarchy"""
    mro = obj_type.mro()
    return list(filter(is_savable_type, mro))



# -*- coding: utf-8 -*-
from typing import List, Iterable, Tuple

from . import records as recordsm

__all__ = 'MergeResult', 'DeleteResult'


class MergeResult:
    """Information about the results from a merge operation.
    `all` contains all of the IDs that were considered in the merge which means not only those that were passed but also
    all those that they reference.
    `merged` contains the ids of all the records that were actually merged (the rest were already present)
     """
    __slots__ = 'all', 'merged'

    def __init__(self, all_snapshots=None, merged_snapshots=None):
        self.all = []  # type: List[recordsm.SnapshotId]
        self.merged = []  # type: List[recordsm.SnapshotId]
        if all_snapshots:
            self.all.extend(all_snapshots)
        if merged_snapshots:
            self.merged.extend(merged_snapshots)

    def update(self, result: 'MergeResult'):
        self.all.extend(result.all)
        self.merged.extend(result.merged)


class DeleteResult:
    """Information about results from a delete operation."""

    __slots__ = '_deleted', '_not_found', '_files_transferred'

    def __init__(self, deleted: list, not_found: Iterable = None):
        self._deleted = tuple(deleted)
        self._not_found = tuple() if not_found is None else tuple(not_found)

    @property
    def deleted(self) -> Tuple:
        return self._deleted

    @property
    def not_found(self) -> Tuple:
        return self._not_found



# -*- coding: utf-8 -*-
import argparse
import datetime

import pytest

import mincepy
from mincepy import expr
from mincepy import fields
from mincepy import saving

# pylint: disable=too-few-public-methods, invalid-name, pointless-statement, protected-access


class Image(fields.WithFields):
    width = fields.field()
    height = fields.field()
    _fmt = mincepy.field(store_as='format')

    def __init__(self, width: int, height: int, fmt: str, creator=None):
        super().__init__()
        self.width = width
        self.height = height
        self._fmt = fmt
        self._creator = creator

    def area(self):
        return self.width * self.height

    @mincepy.field(store_as='_creator')
    def creator(self):
        return self._creator

    @creator.setter
    def creator(self, value):
        self._creator = value


class Profile(fields.WithFields):
    name = fields.field()
    thumbnail = fields.field(type=Image)  # Of type image

    def __init__(self, name, thumbnail):
        self.name = name
        self.thumbnail = thumbnail


def test_field_basics():
    # Can't use full stop in field name
    with pytest.raises(ValueError):
        fields.FieldProperties(store_as='my.field')

    # Try getting an invalid attribute
    field = fields.field(type=Image)
    with pytest.raises(AttributeError):
        field.area

    # However, this is fine with dynamic fields
    field = fields.field(dynamic=True)
    assert isinstance(field.area, fields.Field)
    # The dynamic property should be propagated
    assert isinstance(field.area.height, fields.Field)

    # By default a field is read/write/deletable depending on the object
    field = fields.field(attr='value')
    ns = argparse.Namespace()
    field.__set__(ns, 5)
    assert field.__get__(ns) == 5
    field.__delete__(ns)
    assert 'value' not in ns.__dict__

    # however, if it is converted to a property it becomes read-only until a setter is given
    ns = argparse.Namespace()
    ns.value = 5
    field = fields.field(attr='value')(lambda obj: getattr(obj, 'value'))
    assert field.__get__(ns) == 5
    with pytest.raises(AttributeError):
        field.__set__(ns, 10)
    # But if we add the setter...
    field.setter(lambda obj, value: setattr(obj, 'value', value))
    field.__set__(ns, 10)
    assert field.__get__(ns) == 10

    # Now check deleting
    with pytest.raises(AttributeError):
        field.__delete__(ns)
    field.deleter(lambda obj: delattr(obj, 'value'))
    field.__delete__(ns)
    with pytest.raises(AttributeError):
        field.__get__(ns)

    # Test unreadable attribute
    field = fields.field(attr='unexistent')

    class Test:
        pass

    with pytest.raises(AttributeError):
        field.__get__(Test())


def test_immutable_property():
    """Check that we can handle immutable properties"""

    class Timestamp(fields.WithFields):

        def __init__(self):
            self._creation_time = datetime.datetime.now()

        @fields.field(attr='_creation_time')
        def creation_time(self):
            return self._creation_time

        # No setter for creation time

    ts = Timestamp()
    saved_state = saving.save_instance_state(ts)
    # Now manually fiddle the creation time
    ctime = ts.creation_time
    ts._creation_time = None
    saving.load_instance_state(ts, saved_state)

    # Check that it's been restored correctly
    assert (ts._creation_time == ctime) is True  # pylint: disable=comparison-with-callable
    assert ts.creation_time == ctime


def test_fields():
    # pylint: disable=protected-access
    img = Image(1024, 768, 'png')
    img_dict = saving.save_instance_state(img)
    assert img_dict == dict(
        width=1024,
        height=768,
        format='png',  # Because we asked 'fmt' to be stored as 'format'
        _creator=None,
    )

    img.creator = 'martin'
    img_dict = saving.save_instance_state(img)
    assert img_dict == dict(
        width=1024,
        height=768,
        format='png',  # Because we asked 'fmt' to be stored as 'format'
        _creator='martin',
    )

    img_dict['_creator'] = 'sonia'
    saving.load_instance_state(img, img_dict)
    assert img.creator == 'sonia'

    # Now check what it does if we miss a value
    with pytest.raises(ValueError):
        saving.load_instance_state(img, dict(width=512, height=512), ignore_missing=False)

    # Check that it hasn't destroyed the state
    assert img.width == 1024
    assert img.height == 768
    assert img._fmt == 'png'
    assert img.creator == 'sonia'

    saving.load_instance_state(img, {}, ignore_missing=True)
    assert img.width is None
    assert img.height is None
    assert img._creator is None

    # Check that we can get field using getitem
    width_field = Image['width']
    assert isinstance(width_field, fields.Field)
    assert width_field._properties.attr_name == 'width'


def test_nested_fields():
    assert expr.query_expr(Profile.thumbnail.width == 64) == {'thumbnail.width': 64}

    assert expr.query_expr(Profile.thumbnail == {'width': 64, 'height': 128}) == \
           {'thumbnail': {'width': 64, 'height': 128}}

    # Add a query context
    tag_eq_holiday = fields.field('tag') == 'holiday'
    Profile.thumbnail.set_query_context(tag_eq_holiday)
    # check the query context is being used
    assert expr.query_expr(Profile.thumbnail == {'width': 64, 'height': 128}) == \
           {'$and': [{'tag': 'holiday'}, {'thumbnail': {'width': 64, 'height': 128}}]}

    # now check that it is carried over to 'width'
    assert expr.query_expr(Profile.thumbnail.width == 64) == \
           {'$and': [{'tag': 'holiday'}, {'thumbnail.width': 64}]}


class Animal(fields.WithFields):

    def __init__(self):
        self._specie = None

    def set_specie(self, specie):
        self._specie = specie

    def get_specie(self):
        return self._specie

    def del_specie(self):
        del self._specie

    specie = fields.field()(get_specie, set_specie, del_specie)


def test_field_property():
    seal = Animal()
    seal.specie = 'seal'
    assert seal.specie == 'seal'
    assert saving.save_instance_state(seal) == dict(specie='seal')
    del seal.specie
    with pytest.raises(AttributeError):
        print(seal.specie)

    assert repr(Animal.specie._properties).startswith('FieldProperties(')


def test_with_fields_constructor():
    """Check the default constructor provided by with fields"""

    class Car(fields.WithFields):
        make = fields.field('_make')
        colour = fields.field(default='red')

    car = Car(make='ferrari', colour='red')
    assert car.make == 'ferrari'
    assert car._make == 'ferrari'
    assert car.colour == 'red'

    with pytest.raises(ValueError):
        Car(year=1085)

    # Test default colour
    assert Car(make='honda').colour == 'red'



import mincepy
from mincepy import testing


def test_db():
    """Test using the db() free-function to get type helpers"""
    helper = mincepy.db(testing.Car)
    assert isinstance(helper, mincepy.TypeHelper)
    assert helper.TYPE_ID == testing.Car.TYPE_ID



# -*- coding: utf-8 -*-
import pathlib
import shutil
import tempfile
from typing import Optional, BinaryIO

from . import type_ids
from . import base_savable
from . import fields

__all__ = 'File', 'BaseFile'


class File(base_savable.SimpleSavable):
    """A mincePy file object.  These should not be instantiated directly but using Historian.create_file()"""

    TYPE_ID = type_ids.FILE_TYPE_ID
    READ_SIZE = 256  # The number of bytes to read at a time

    def __init__(self, file_store, filename: str = None, encoding=None):
        super().__init__()
        self._file_store = file_store
        self._filename = filename
        self._encoding = encoding
        self._file_id = None
        self._buffer_file = _create_buffer_file()

    @fields.field('_filename')
    def filename(self) -> Optional[str]:
        return self._filename

    @fields.field('_encoding')
    def encoding(self) -> Optional[str]:
        return self._encoding

    @fields.field('_file_id')
    def file_id(self):
        return self._file_id

    def open(self, mode='r', **kwargs) -> BinaryIO:
        """Open returning a file like object that supports close() and read()"""
        self._ensure_buffer()
        if 'b' not in mode:
            kwargs.setdefault('encoding', self.encoding)
        return open(self._buffer_file, mode, **kwargs)

    def from_disk(self, path):
        """Copy the contents of a disk file to this file"""
        with open(str(path), 'r', encoding=self.encoding) as disk_file:
            with self.open('w') as this:
                shutil.copyfileobj(disk_file, this)

    def to_disk(self, folder: [str, pathlib.Path]):
        """Copy the contents of this file to a file on disk in the given folder"""
        file_path = pathlib.Path(str(folder)) / self.filename
        with open(str(file_path), 'w', encoding=self._encoding) as disk_file:
            with self.open('r') as this:
                shutil.copyfileobj(this, disk_file)

    def write_text(self, text: str, encoding=None):
        encoding = encoding or self._encoding
        with self.open('w', encoding=encoding) as fileobj:
            fileobj.write(text)

    def read_text(self, encoding=None) -> str:
        """Read the contents of the file as text.
        This function is named as to mirror pathlib.Path"""
        encoding = encoding or self._encoding
        with self.open('r', encoding=encoding) as fileobj:
            return fileobj.read()

    def save_instance_state(self, saver):
        filename = self.filename or ''
        with open(self._buffer_file, 'rb') as fstream:
            self._file_id = self._file_store.upload_from_stream(filename, fstream)

        return super().save_instance_state(saver)

    def load_instance_state(self, saved_state, loader):
        super().load_instance_state(saved_state, loader)
        self._file_store = loader.get_archive().file_store
        # Don't copy the file over now, do it lazily when the file is first opened
        self._buffer_file = None

    def _ensure_buffer(self):
        if self._buffer_file is None:
            if self._file_id is not None:
                self._update_buffer()
            else:
                _create_buffer_file()

    def _update_buffer(self):
        self._buffer_file = _create_buffer_file()
        with open(self._buffer_file, 'wb') as fstream:
            self._file_store.download_to_stream(self._file_id, fstream)

    def __str__(self):
        contents = [str(self._filename)]
        if self._encoding is not None:
            contents.append('({})'.format(self._encoding))
        return ' '.join(contents)

    def __eq__(self, other) -> bool:
        """Compare the contents of two files

        If both files do not exist they are considered equal.
        """
        if not isinstance(other, File) or self.filename != other.filename:  # pylint: disable=comparison-with-callable
            return False

        try:
            with self.open() as my_file:
                try:
                    with other.open() as other_file:
                        while True:
                            my_line = my_file.readline(self.READ_SIZE)
                            other_line = other_file.readline(self.READ_SIZE)
                            if my_line != other_line:
                                return False
                            if my_line == '' and other_line == '':
                                return True
                except FileNotFoundError:
                    return False
        except FileNotFoundError:
            # Our file doesn't exist, make sure the other doesn't either
            try:
                with other.open():
                    return False
            except FileNotFoundError:
                return True

    def yield_hashables(self, hasher):
        """Hash the contents of the file"""
        try:
            with self.open('rb') as opened:
                while True:
                    line = opened.read(self.READ_SIZE)
                    if line == b'':
                        return
                    yield line
        except FileNotFoundError:
            yield from hasher.yield_hashables(None)


def _create_buffer_file():
    tmp_file = tempfile.NamedTemporaryFile(delete=False)
    tmp_path = tmp_file.name
    tmp_file.close()
    return tmp_path


BaseFile = File  # Here just for legacy reasons.  Deprecate in 1.0



# -*- coding: utf-8 -*-
import pytest

import mincepy
from mincepy import testing
from mincepy.testing import Car

# pylint: disable=protected-access


def test_transaction_snapshots(historian: mincepy.Historian):
    ferrari = Car('ferrari')
    historian.save(ferrari)
    ferrari_sid = historian.get_snapshot_id(ferrari)

    with historian.transaction():
        ferrari_snapshot_1 = historian.load_snapshot(ferrari_sid)
        with historian.transaction():
            ferrari_snapshot_2 = historian.load_snapshot(ferrari_sid)
            # Reference wise they should be unequal
            assert ferrari_snapshot_1 is not ferrari_snapshot_2
            assert ferrari is not ferrari_snapshot_1
            assert ferrari is not ferrari_snapshot_2

            # Value wise they should be equal
            assert ferrari == ferrari_snapshot_1
            assert ferrari == ferrari_snapshot_2

        # Now check within the same transaction the result is the same
        ferrari_snapshot_2 = historian.load_snapshot(ferrari_sid)
        # Reference wise they should be unequal
        assert ferrari_snapshot_1 is not ferrari_snapshot_2
        assert ferrari is not ferrari_snapshot_1
        assert ferrari is not ferrari_snapshot_2

        # Value wise they should be equal
        assert ferrari == ferrari_snapshot_1
        assert ferrari == ferrari_snapshot_2


def test_transaction_records(historian: mincepy.Historian):
    """Make sure that records within a transaction are not recreated at each save"""
    with historian.transaction():
        ferrari = Car('ferrari')

        # Save and get the record for the ferrari
        ferrari_id = historian.save(ferrari)
        record = historian.get_current_record(ferrari)

        ferrari_id2 = historian.save(ferrari)
        assert ferrari_id == ferrari_id2
        assert historian.get_current_record(ferrari) is record

        loaded = historian.load(ferrari_id)
        assert loaded is ferrari


def test_find(historian: mincepy.Historian):
    honda_id = historian.save(Car('honda'))
    zonda_id = historian.save(Car('zonda'))
    porsche_id = historian.save(Car('porsche'))

    cars = list(historian.find(Car))
    assert len(cars) == 3

    makes = [car.make for car in cars]
    assert 'honda' in makes
    assert 'zonda' in makes
    assert 'porsche' in makes

    obj_ids = [car.obj_id for car in cars]
    assert honda_id in obj_ids
    assert zonda_id in obj_ids
    assert porsche_id in obj_ids


def test_sync(historian: mincepy.Historian):
    car = Car('ferrari', 'red')
    historian.save_one(car)

    # Simulate saving the car from another connection
    honda = Car('honda', 'black')
    historian.save_one(honda)
    honda_record = historian.get_current_record(honda)

    archive = historian.archive
    record = historian.get_current_record(car)
    builder = record.child_builder(obj_id=historian.get_obj_id(car),
                                   snapshot_hash=honda_record.snapshot_hash,
                                   state=honda_record.state,
                                   state_types=honda_record.state_types)
    archive.save(builder.build())

    # Now update and check the state
    historian.sync(car)
    assert car.make == 'honda'
    assert car.colour == 'black'

    # Also, check the cached record
    car_record = historian.get_current_record(car)
    assert car_record.snapshot_hash == honda_record.snapshot_hash
    assert car_record.state == honda_record.state

    # Check that syncing an unsaved object returns False (because there's nothing to do)
    assert historian.sync(Car()) is False


def test_to_obj_id(historian: mincepy.Historian):
    car = Car()
    car_id = car.save()

    assert historian.to_obj_id(car_id) is car_id
    assert historian.to_obj_id(car) == car_id
    assert historian.to_obj_id(str(car_id)) == car_id
    assert historian.to_obj_id('carrot') is None


def test_copy(historian: mincepy.Historian):
    car = Car('zonda')

    historian.save(car)
    car_copy = mincepy.copy(car)
    assert car == car_copy
    assert car is not car_copy
    car_copy.save()

    record = historian.get_current_record(car)
    copy_record = historian.get_current_record(car_copy)

    assert record is not copy_record
    assert copy_record.get_copied_from() == record.snapshot_id


def test_copy_unsaved(historian: mincepy.Historian):
    car = Car('porsche', 'silver')
    car_copy = mincepy.copy(car)

    assert car_copy is not car
    assert car == car_copy

    # The cars should not be saved
    assert not historian.is_known(car)
    assert not historian.is_known(car_copy)


def test_save(historian: mincepy.Historian):
    car = Car('porsche', 'yellow')
    with pytest.raises(ValueError):
        historian.save((car, {'speed': 'fast'}, 124))


def test_is_trackable(historian: mincepy.Historian):
    assert historian.is_trackable(mincepy.testing.Car) is True
    assert historian.is_trackable(5) is False
    assert historian.is_trackable(5.6) is False
    assert historian.is_trackable('hello') is False
    assert historian.is_trackable(False) is False
    assert historian.is_trackable(b'byte me') is False


def test_delete_referenced(historian: mincepy.Historian):
    car = testing.Car()
    garage = testing.Garage(mincepy.ObjRef(car))
    garage.save()

    # Should be prevented from deleting the car as it is referenced by the garage
    try:
        historian.delete(car)
    except mincepy.ReferenceError as exc:
        assert exc.references == {car.obj_id}
    else:
        assert False, 'Reference error should have been raised'

    historian.delete(garage)
    # Now safe to delete car
    historian.delete(car)

    car = testing.Car()
    garage = testing.Garage(mincepy.ObjRef(car))
    garage.save()

    # Now, check that deleting both together works
    historian.delete(car, garage)


def test_find_arg_types(historian: mincepy.Historian):
    """Test the argument types accepted by the historian find() method"""
    red_ferrari = testing.Car(colour='red', make='ferrari')
    green_ferrari = testing.Car(colour='green', make='ferrari')
    red_honda = testing.Car(colour='red', make='honda')
    martin = testing.Person(name='martin', age=35, car=red_honda)

    red_ferrari_id, green_ferrari_id, red_honda_id = \
        historian.save(red_ferrari, green_ferrari, red_honda)
    martin_id = martin.save()

    # Test different possibilities for object ids being passed
    list(historian.find(obj_id=red_ferrari_id))
    list(historian.find(obj_id=[red_ferrari_id, green_ferrari_id, martin_id, red_honda_id]))
    list(historian.find(obj_id=(red_ferrari_id, green_ferrari_id, martin_id, red_honda_id)))
    list(historian.find(obj_id=str(red_ferrari_id)))

    # Test object types
    list(historian.find(obj_type=testing.Person))
    list(historian.find(obj_type=[testing.Person, testing.Car]))
    list(historian.find(obj_type=(testing.Person, testing.Car)))
    list(historian.find(obj_type=testing.Person.TYPE_ID))
    list(historian.find(obj_type=[testing.Person.TYPE_ID, testing.Car.TYPE_ID]))


def test_concurrent_modification(historian: mincepy.Historian, archive_uri: str):
    # Create a second historian connected to the same archive
    historian2 = mincepy.connect(archive_uri, use_globally=False)
    ferrari = testing.Car(colour='red', make='ferrari')
    ferrari_id = historian.save(ferrari)
    ferrari2 = historian2.load(ferrari_id)

    assert ferrari_id == ferrari2.obj_id
    assert ferrari is not ferrari2, \
        "The archive don't know about each other so the objects instances should not be the same"

    # Repaint
    ferrari.colour = 'yellow'
    historian.save(ferrari)

    # Now change ferrari2 and see what happens
    ferrari2.colour = 'green'
    with pytest.raises(mincepy.ModificationError):
        historian2.save(ferrari2)

    # Now, let's sync up
    assert historian2.sync(ferrari2), "ferrari2 hasn't been updated"
    assert ferrari2.colour == 'yellow'


def test_replace_simple(historian: mincepy.Historian):

    def paint_shop(car, colour):
        """An imaginary function that modifies an object but returns a copy rather than an in
        place modification"""
        return Car(car.make, colour)

    honda = Car('honda', 'yellow')
    honda_id = historian.save(honda)

    # Now paint the honda
    new_honda = paint_shop(honda, 'green')
    assert historian.get_obj_id(honda) == honda_id

    # Now we know that this is a 'continuation' of the history of the original honda, so replace
    historian.replace(honda, new_honda)
    assert historian.get_obj_id(honda) is None

    assert historian.get_obj_id(new_honda) == honda_id
    historian.save(new_honda)
    del honda, new_honda

    loaded = historian.load(honda_id)
    assert loaded.make == 'honda'
    assert loaded.colour == 'green'

    with pytest.raises(RuntimeError):
        # Check that we can't replace in a transaction
        with historian.transaction():
            historian.replace(loaded, Car())


def test_snapshots_collection(historian: mincepy.Historian):
    ferrari = testing.Car(colour='red', make='ferrari')
    ferrari_id = ferrari.save()

    records = list(historian.snapshots.records.find())
    assert len(records) == 1

    snapshots = list(historian.snapshots.find())
    assert len(snapshots) == 1
    assert snapshots[0] == ferrari

    ferrari.colour = 'brown'
    ferrari.save()

    records = list(historian.snapshots.records.find())
    assert len(records) == 2

    snapshots = list(historian.snapshots.find())
    assert len(snapshots) == 2
    assert set(car.colour for car in snapshots) == {'red', 'brown'}

    assert historian.snapshots.records.find(Car.colour == 'brown',
                                            obj_id=ferrari_id).one().version == 1


def test_objects_collection(historian: mincepy.Historian):
    ferrari = testing.Car(colour='red', make='ferrari')
    ferrari_id = ferrari.save()

    records = list(historian.objects.records.find())
    assert len(records) == 1

    objects = list(historian.objects.find())
    assert len(objects) == 1
    assert objects[0] is ferrari

    ferrari.colour = 'brown'
    ferrari.save()

    records = list(historian.objects.records.find())
    assert len(records) == 1

    objects = list(historian.objects.find())
    assert len(objects) == 1
    assert set(car.colour for car in objects) == {'brown'}

    assert historian.objects.records.find(Car.colour == 'brown',
                                          obj_id=ferrari_id).one().version == 1


def test_get_obj_type(historian: mincepy.Historian):
    assert historian.get_obj_type(Car.TYPE_ID) is Car


def test_get_obj_id(historian: mincepy.Historian):
    """Test the get_obj_id method"""
    unsaved = testing.Car()
    car = testing.Car()
    obj_id = car.save()

    assert historian.get_obj_id(car) is obj_id
    assert historian.get_obj_id(unsaved) is None
    with historian.transaction():
        assert historian.get_obj_id(car) is obj_id
        historian.delete(car)
        assert historian.get_obj_id(car) is None
    assert historian.get_obj_id(car) is None


def test_merge(historian: mincepy.Historian):
    with testing.temporary_historian(
            testing.create_archive_uri(db_name='test_historian')) as remote:
        local = historian
        # remote = clean_test_historian

        remote_skoda = testing.Car(make='skoda', colour='green')
        skoda_id = remote.save(remote_skoda)
        assert remote_skoda._historian is remote

        result = local.merge(remote.objects.find(obj_id=skoda_id))
        assert remote.get_snapshot_id(remote_skoda) in result.merged
        assert local.find(obj_id=skoda_id).count() == 1

        # Now, let's update and see if we can merge
        remote_skoda.colour = 'yellow'
        remote.save(remote_skoda)

        result = local.merge(remote.objects.find(obj_id=skoda_id))
        assert remote.get_snapshot_id(remote_skoda) in result.merged
        assert local.snapshots.find(obj_id=skoda_id).count() == 2
        assert local.find(obj_id=skoda_id).one().colour == 'yellow'

        # Now, change both to the same thing
        local_skoda = local.load(skoda_id)
        assert local_skoda._historian is local
        assert local_skoda is not remote_skoda

        local_skoda.colour = 'blue'
        local_skoda.save()

        remote_skoda.colour = 'blue'
        remote_skoda.save()
        result = local.merge(remote.objects.find(obj_id=skoda_id))
        assert not result.merged  # None should have been transferred

        # Now check that conflicts are correctly handled
        remote_skoda.colour = 'brown'
        remote_skoda.save()
        local_skoda.colour = 'grey'
        local_skoda.save()
        with pytest.raises(mincepy.MergeError):
            local.merge(remote.objects.find(obj_id=skoda_id))


def test_large_merge(
        historian: mincepy.Historian,
        large_dataset,  # pylint: disable=unused-argument
):
    with testing.temporary_historian(
            testing.create_archive_uri(db_name='test_historian')) as remote:
        local = historian

        all_objects = local.find()
        assert all_objects.count() > 0

        # Merge the large dataset into the remote historian
        result = remote.merge(all_objects)
        assert len(result.all) == len(result.merged) == local.find().count()
        assert local.find().count() == remote.find().count()


def test_merge_file(historian: mincepy.Historian):
    """Test that merging files works correctly"""
    file = historian.create_file('test.dat')
    file.write_text('bla bla')
    file.save()
    with testing.temporary_historian(
            testing.create_archive_uri(db_name='test_historian')) as remote:
        local = historian
        # Merge the file into the remote
        result = remote.merge(local.find(obj_id=file.obj_id))
        assert len(result.all) == len(result.merged) == local.find().count()
        assert local.find().count() == remote.find().count()

        remote_file = remote.get(file.obj_id)
        assert file.read_text() == remote_file.read_text()

        # Now check that files contained within objects are correctly merged
        file_list = mincepy.List((file,))
        file_list.save()
        result = remote.merge(local.find(obj_id=file_list.obj_id))
        assert len(result.merged) == 1
        assert historian.get_snapshot_id(file_list) in result.merged

        remote_file_list = remote.get(file_list.obj_id)
        assert file.read_text() == remote_file_list[0].read_text()



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
import collections
import contextlib

try:
    from contextlib import nullcontext
except ImportError:
    from contextlib2 import nullcontext
import getpass
import logging
import socket
import typing
from typing import MutableMapping, Any, Optional, Iterable, Union, Iterator, Type, Dict, Callable
import weakref

import deprecation
import networkx

from . import archives
from . import builtins
from . import frontend
from . import defaults
from . import depositors
from . import refs
from . import exceptions
from . import expr
from . import files
from . import helpers
from . import hist
from . import migrate
from . import operations
from . import qops
from . import records as recordsm  # The records module
from . import result_types
from . import staging
from . import tracking
from . import types
from . import type_registry
from . import utils
from . import version as version_mod
from .transactions import RollbackTransaction, Transaction, LiveObjects

__all__ = 'Historian', 'ObjectEntry'

logger = logging.getLogger(__name__)

ObjectEntry = collections.namedtuple('ObjectEntry', 'ref obj')
HistorianType = Union[helpers.TypeHelper, typing.Type[types.SavableObject]]


class Historian:  # pylint: disable=too-many-public-methods, too-many-instance-attributes
    """The historian acts as a go-between between your python objects and the archive which is
    a persistent store of the records.  It will keep track of all live objects (i.e. those that
    have active references to them) that have been loaded and/or saved as well as enabling the
    user to lookup objects in the archive."""

    @deprecation.deprecated(deprecated_in='0.14.5',
                            removed_in='0.16.0',
                            current_version=version_mod.__version__,
                            details='Use mincepy.copy() instead')
    def copy(self, obj):  # pylint: disable=no-self-use
        """Create a shallow copy of the object.  Using this method allows the historian to inject
        information about where the object was copied from into the record if saved."""
        return tracking.copy(obj)

    @deprecation.deprecated(deprecated_in='0.15.10',
                            removed_in='0.17.0',
                            current_version=version_mod.__version__,
                            details='Use mincepy.records.find() instead')
    def find_records(self, *args, **kwargs) -> Iterator[recordsm.DataRecord]:
        """Find records

        Has same signature as py:meth:`mincepy.Records.find`.
        """
        yield from self.records.find(*args, **kwargs)

    @deprecation.deprecated(deprecated_in='0.15.10',
                            removed_in='0.17.0',
                            current_version=version_mod.__version__,
                            details='Use mincepy.records.distinct() instead')
    def find_distinct(self, *args, **kwargs):
        """Get distinct values of the given record key

        Has same signature as py:meth:`mincepy.Records.distinct`.
        """
        yield from self.records.distinct(*args, **kwargs)

    def __init__(self, archive: archives.Archive, equators=()):
        self._archive = archive
        self._equator = types.Equator(defaults.get_default_equators() + equators)
        # Register default types
        self._type_registry = type_registry.TypeRegistry()
        self.register_type(refs.ObjRef)
        self.register_types(builtins.HISTORIAN_TYPES)
        self.register_type(builtins.SnapshotIdHelper())
        self.register_types(archive.get_types())

        # Snapshot objects -> reference. Objects that were loaded from historical snapshots
        self._snapshots_objects = utils.WeakObjectIdDict(
        )  # type: MutableMapping[Any, recordsm.SnapshotId]
        self._live_objects = LiveObjects()

        self._transactions = None

        self._saving_set = set()

        self._user = getpass.getuser()
        self._hostname = socket.gethostname()

        self._live_depositor = depositors.LiveDepositor(self)
        self._meta = hist.Meta(self, self._archive)
        self._migrate = migrate.Migrations(self)
        self._references = hist.References(self)

        self._snapshots = frontend.ObjectCollection(
            self,
            self._archive.snapshots,
            record_factory=lambda record_dict: SnapshotLoadableRecord(
                record_dict, self.load_snapshot_from_record),
            obj_loader=self.load_snapshot_from_record)
        self._objects = frontend.ObjectCollection(
            self,
            self._archive.objects,
            record_factory=lambda record_dict: LoadableRecord(
                record_dict, self.load_snapshot_from_record, self._load_object_from_record),
            obj_loader=self._load_object_from_record)

    @property
    def archive(self):
        return self._archive

    @property
    def meta(self) -> hist.Meta:
        """Access to functions that operate on the metadata"""
        return self._meta

    @property
    def primitives(self) -> tuple:
        """A tuple of all the primitive types"""
        return types.PRIMITIVE_TYPES + (self._archive.get_id_type(),)

    @property
    def migrations(self) -> migrate.Migrations:
        """Access the migration possibilities"""
        return self._migrate

    @property
    def records(self) -> frontend.EntriesCollection['LoadableRecord']:
        """Access methods and properties that act on and return data records"""
        return self._objects.records

    @property
    def objects(self) -> frontend.ObjectCollection:
        """Access the snapshots"""
        return self._objects

    @property
    def references(self) -> hist.References:
        """Access the references possibilities"""
        return self._references

    @property
    def snapshots(self) -> frontend.ObjectCollection:
        """Access the snapshots"""
        return self._snapshots

    def create_file(self, filename: str = None, encoding: str = None) -> builtins.BaseFile:
        """Create a new file.  The historian will supply file type compatible with the archive in
         use."""
        return files.File(self._archive.file_store, filename, encoding)

    def save(self, *objs: object):
        """Save multiple objects producing corresponding object identifiers.  This returns a
        sequence of ids that is in the same order as the passed objects.

        :param objs: the object(s) to save.  Can also be a tuple of (obj, meta) to optionally
            include metadata to be saved with the object(s)
        """
        to_save = []
        # Convert everything to tuples
        for entry in objs:
            if isinstance(entry, tuple):
                if len(entry) > 2:
                    raise ValueError(
                        "Supplied tuples can only contain (object, meta), got '{}'".format(entry))
            else:
                entry = (entry,)
            to_save.append(entry)

        ids = []
        with self.in_transaction():
            for entry in to_save:
                ids.append(self.save_one(*entry))

        if len(objs) == 1:
            return ids[0]

        return ids

    def save_one(self, obj: object, meta: dict = None):
        """Save the object returning an object id.  If metadata is supplied it will be set on the
        object.

        Developer note: this is the front end point-of-entry for a user/client code saving an object
        however subsequent objects being saved in this transaction will only go through _save_object
        and therefore any code common to all objects being saved should possibly go there.
        """
        if obj in self._snapshots_objects:
            raise exceptions.ModificationError(
                'Cannot save a snapshot object, that would rewrite history!')

        if meta and not isinstance(meta, dict):
            raise TypeError("Metadata must be a dictionary, got type '{}'".format(type(meta)))

        # Save the object and metadata
        with self.in_transaction():
            record = self._save_object(obj, self._live_depositor)
            if meta:
                self.meta.update(record.obj_id, meta)

        return record.obj_id

    def is_known(self, obj: object) -> bool:
        """Check if an object has ever been saved and is therefore known to the historian

        :return: True if ever saved, False otherwise
        """
        return self.get_obj_id(obj) is not None

    def replace(self, old: object, new: object):
        """Replace a live object with a new version.

        This is especially useful if you have made a copy of an object and modified it but you want
        to continue the history of the object as the original rather than a brand new object.  Then
        just replace the old object with the new one by calling this function.
        """
        if self.current_transaction() is not None:
            raise RuntimeError("Can't replace during a transaction for the time being")

        if not isinstance(new, type(old)):
            raise TypeError("Can't replace type '{} with type '{}!".format(type(old), type(new)))

        # Get the current record and replace the object with the new one
        record = self._live_objects.get_record(old)
        self._live_objects.remove(record.obj_id)
        self._live_objects.insert(new, record)

        # Make sure creators is correct as well
        staging.replace(old, new)

    def load_snapshot(self, snapshot_id: recordsm.SnapshotId) -> object:
        return self._new_snapshot_depositor().load(snapshot_id)

    def load_snapshot_from_record(self, record: recordsm.DataRecord) -> object:
        return self._new_snapshot_depositor().load_from_record(record)

    def load(self, *obj_id_or_snapshot_id):
        """Load object(s) or snapshot(s)."""
        loaded = []
        for entry in obj_id_or_snapshot_id:
            loaded.append(self.load_one(entry))

        if len(obj_id_or_snapshot_id) == 1:
            return loaded[0]

        return loaded

    def load_one(self, obj_id_or_snapshot_id):
        """Load one object or snapshot from the database"""
        if isinstance(obj_id_or_snapshot_id, recordsm.SnapshotId):
            return self.load_snapshot(obj_id_or_snapshot_id)

        return self._load_object(obj_id_or_snapshot_id, self._live_depositor)

    def get(self, obj_id) -> object:
        """Get a live object using the object id"""
        return self._objects.get(obj_id)

    def sync(self, obj: object) -> bool:
        """Update an object with the latest state in the database.
        If there is no new version in the archive then the current version remains
        unchanged including any modifications.

        :return: True if the object was updated, False otherwise
        """
        obj_id = self.get_obj_id(obj)
        if obj_id is None:
            # Never saved so the object is as up to date as can be!
            return False

        record = self._objects.records.get(obj_id)

        if record.is_deleted_record():
            raise exceptions.ObjectDeleted("Object with id '{}' has been deleted".format(obj_id))

        if record.version == self.get_snapshot_id(obj).version:
            # Nothing has changed
            return False

        # The one in the archive is newer, so use that
        return self._live_depositor.update_from_record(obj, record)

    def delete(self, *obj_or_identifier, imperative=True) -> result_types.DeleteResult:
        """Delete an object.

        :param imperative: if True, this means that the caller explicitly expects this call to delete the passed objects
            and it should therefore raise if an object cannot be found or has been deleted already.  If False, the
            function will ignore these cases and continue.
        :raises mincepy.NotFound: if the object cannot be found (potentially because it was
            already deleted)
        """
        # We need the current records to be able to build the delete records
        obj_ids = list(map(self._ensure_obj_id, obj_or_identifier))

        # Find the current records
        records = {}  # type: Dict[Any, recordsm.DataRecord]
        left_to_find = set()
        for obj_id in obj_ids:
            try:
                records[obj_id] = self.get_current_record(self.get_obj(obj_id))
            except exceptions.ObjectDeleted:
                if imperative:
                    # Object deleted already so reraise
                    raise
            except exceptions.NotFound:
                left_to_find.add(obj_id)

        # Those that we don't have cached records for and need to look up
        if left_to_find:
            # Have a look in the archive
            for record in self._objects.records.find(recordsm.DataRecord.obj_id.in_(*left_to_find)):
                records[record.obj_id] = record
                left_to_find.remove(record.obj_id)

        if left_to_find and imperative:
            raise exceptions.NotFound(left_to_find)

        deleted = []
        with self.in_transaction() as trans:
            # Mark each object as deleted in the transaction and stage the delete record for insertion
            # in the order that they were passed to us, in case this makes a difference to the caller
            for obj_id in obj_ids:
                record = records.get(obj_id, None)
                if record is None:
                    continue

                builder = recordsm.make_deleted_builder(record)
                deleted_record = self._record_builder_created(builder).build()
                trans.delete(record.obj_id)
                trans.stage(operations.Insert(deleted_record))
                deleted.append(record.obj_id)

        return result_types.DeleteResult(deleted, left_to_find)

    def history(
            self,
            obj_or_obj_id,
            idx_or_slice='*',
            as_objects=True
    ) -> [typing.Sequence[ObjectEntry], typing.Sequence[recordsm.DataRecord]]:
        """Get a sequence of object ids and instances from the history of the given object.

        :param obj_or_obj_id: The instance or id of the object to get the history for
        :param idx_or_slice: The particular index or a slice of which historical versions to get
        :param as_objects: if True return the object instances, otherwise returns the DataRecords

        Example:
        >>> import mincepy, mincepy.testing
        >>> historian = mincepy.get_historian()
        >>> car = mincepy.testing.Car('ferrari', 'white')
        >>> car_id = historian.save(car)
        >>> car.colour = 'red'
        >>> historian.save(car)
        >>> history = historian.history(car_id)
        >>> len(history)
        2
        >>> history[0].obj.colour == 'white'
        True
        >>> history[1].obj.colour == 'red'
        True
        >>> history[1].obj is car
        """
        obj_id = self._ensure_obj_id(obj_or_obj_id)
        snapshot_ids = self._archive.get_snapshot_ids(obj_id)
        indices = utils.to_slice(idx_or_slice)
        to_get = snapshot_ids[indices]
        if as_objects:
            return [ObjectEntry(sid, self.load(sid)) for sid in to_get]

        return [self._archive.load(ref) for ref in to_get]

    def get_current_record(self, obj: object) -> recordsm.DataRecord:
        """Get the current record that the historian has cached for the passed object"""
        trans = self.current_transaction()
        # Try the transaction first
        if trans:
            try:
                return trans.get_record_for_live_object(obj)
            except exceptions.ObjectDeleted:
                # ObjectDeleted is a specialisation of the NotFound error but it means that we
                # should consider the object as being gone so reraise it
                raise
            except exceptions.NotFound:
                pass

        return self._live_objects.get_record(obj)

    def get_obj_id(self, obj: object) -> Any:
        """Get the object ID for a live object.

        :return: the object id or None if the object is not known to the historian
        """
        trans = self.current_transaction()
        if trans is not None:
            try:
                return trans.get_snapshot_id_for_live_object(obj).obj_id
            except exceptions.NotFound:
                pass

        try:
            obj_id = self._live_objects.get_record(obj).obj_id
            if trans is not None and trans.is_deleted(obj_id):
                # The object has been deleted in the transaction, so it is not known
                return None

            return obj_id
        except exceptions.NotFound:
            return None

    def get_obj(self, obj_id) -> object:
        """Get a currently live object"""
        trans = self.current_transaction()
        if trans:
            try:
                return trans.get_live_object(obj_id)
            except exceptions.ObjectDeleted:
                # ObjectDeleted is a specialisation of the NotFound error but it means that we
                # should consider the object as being gone so reraise it
                raise
            except exceptions.NotFound:
                pass

        return self._live_objects.get_object(obj_id)

    def is_saved(self, obj: object) -> bool:
        """Test if an object is saved with this historian.  This is equivalent to
        `historian.get_obj_id(obj) is not None`."""
        return self.get_obj_id(obj) is not None

    def to_obj_id(self, obj_or_identifier):
        """
        This call will try and get an object id from the passed parameter.  The possibilities are:

        1. Passed an object ID in which case it will be returned unchanged
        2. Passed a snapshot ID, in which case the corresponding object ID will be returned
        2. Passed a live object instance, in which case the id of that object will be returned
        3. Passed a type that can be understood by the archive as an object id e.g. a string of
           version, in which case the archive will attempt to convert it

        Returns None if neither of these cases were True.
        """
        if self.is_obj_id(obj_or_identifier):
            return obj_or_identifier

        if isinstance(obj_or_identifier, recordsm.SnapshotId):
            return obj_or_identifier.obj_id

        try:
            # Try creating it for the user by calling the constructor with the argument passed.
            # This helps for common obj id types which can be constructed from a string
            return self._archive.construct_archive_id(obj_or_identifier)
        except (ValueError, TypeError):
            # Maybe we've been passed an object
            pass

        try:
            return self.get_obj_id(obj_or_identifier)
        except exceptions.NotFound:
            pass

        # Out of options
        return None

    def get_snapshot_id(self, obj: object) -> recordsm.SnapshotId:
        """Get the current snapshot id for a live object.  Will return the id or raise
        :class:`mincepy.NotFound` exception"""
        trans = self.current_transaction()
        if trans:
            try:
                return trans.get_snapshot_id_for_live_object(obj)
            except exceptions.NotFound:
                pass

        return self._live_objects.get_record(obj).snapshot_id

    def hash(self, obj: object):
        return self._equator.hash(obj)

    def eq(self, one: object, other: object):  # pylint: disable=invalid-name
        return self._equator.eq(one, other)

    # region Types

    @classmethod
    def is_trackable(cls, obj):
        """Determine if an object is trackable i.e. we can treat these as live objects and
        automatically keep track of their history when saving.  Ultimately this is determined by
        whether the type is weak referencable or not.
        """
        try:
            weakref.ref(obj)
            return True
        except TypeError:
            return False

    @property
    def type_registry(self) -> type_registry.TypeRegistry:
        return self._type_registry

    def is_primitive(self, obj) -> bool:
        """Check if the object is one of the primitives and should be saved by value in the
        archive"""
        return isinstance(obj, self.primitives)

    def is_obj_id(self, obj_id) -> bool:
        """Check if an object is of the object id type"""
        return isinstance(obj_id, self._archive.get_id_type())

    def register_type(self, obj_class_or_helper: HistorianType) -> helpers.TypeHelper:
        helper = self._type_registry.register_type(obj_class_or_helper)
        self._equator.add_equator(helper)
        return helper

    def register_types(self, obj_clases_or_helpers: Iterable[HistorianType]):
        for item in obj_clases_or_helpers:
            self.register_type(item)

    def get_obj_type_id(self, obj_type):
        return self._type_registry.get_type_id(obj_type)

    def get_obj_type(self, type_id):
        return self.get_helper(type_id).TYPE

    def get_helper(self, type_id_or_type, auto_register=False) -> helpers.TypeHelper:
        if auto_register and issubclass(type_id_or_type, types.SavableObject):
            self._ensure_compatible(type_id_or_type)

        return self._type_registry.get_helper(type_id_or_type)

    # endregion

    # pylint: disable=redefined-builtin
    def find(self,
             *filter: expr.FilterSpec,
             obj_type=None,
             obj_id=None,
             version: int = -1,
             state=None,
             meta: dict = None,
             sort=None,
             limit=0,
             skip=0) -> frontend.ResultSet[object]:
        """
        .. _MongoDB: https://docs.mongofrontend.com/manual/tutorial/query-documents/

        Find objects.  This call will search the archive for objects matching the given criteria.
        In many cases the main arguments of interest will be `state` and `meta` which allow you to
        apply filters on the stored state of the object and metadata respectively.  To understand
        how the state is stored in the database (and therefore how to apply filters to it) it may
        be necessary to look at the details of the `save_instance_state()` method for that type.
        Metadata is always a dictionary containing primitives (strings, dicts, lists, etc).

        For the most part, the filter syntax of `mincePy` conforms to that of `MongoDB`_ with
        convenience functions locate in :py:mod:`mincepy.qops` that can make it easier to
        to build a query.

        Examples:

        Find all :py:class:`~mincepy.testing.Car`s that are brown or red:

        >>> import mincepy as mpy
        >>> historian = mpy.get_historian()
        >>> historian.find(mpy.testing.Car.colour.in_('brown', 'red'))

        Find all people that are older than 34 and live in Edinburgh:

        >>> historian.find(mpy.testing.Person.age > 34, meta=dict(city='Edinburgh'))

        :param obj_type: the object type to look for
        :param obj_id: an object or multiple object ids to look for
        :param version: the version of the object to retrieve, -1 means latest
        :param state: the criteria on the state of the object to apply
        :type state: must be subclass of historian.primitive
        :param meta: the search criteria to apply on the metadata of the object
        :param sort: the sort criteria
        :param limit: the maximum number of results to return, 0 means unlimited
        :param skip: the page to get results from
        """
        return self._objects.find(
            *filter,
            obj_type=obj_type,
            obj_id=obj_id,
            version=version,
            state=state,
            meta=meta,
            sort=sort,
            limit=limit,
            skip=skip,
        )

    def get_creator(self, obj_or_identifier) -> object:
        """Get the object that created the passed object"""
        if not self.is_obj_id(obj_or_identifier):
            # Object instance, try the staging area
            info = staging.get_info(obj_or_identifier, create=False) or {}
            created_by = info.get(recordsm.ExtraKeys.CREATED_BY, None)
            if created_by is not None:
                return created_by

        creator_id = self.created_by(obj_or_identifier)
        return self.load_one(creator_id)

    def created_by(self, obj_or_identifier):
        """Return the id of the object that created the passed object"""
        try:
            record = self.get_current_record(obj_or_identifier)
        except exceptions.NotFound as exc:
            if not self.is_obj_id(obj_or_identifier):
                raise

            try:
                record = self._objects.records.find(obj_id=obj_or_identifier).one()
            except exceptions.NotOneError:
                raise exc from None

        return record.created_by

    def get_user_info(self) -> dict:
        """Get information about the current user and host"""
        user_info = {}
        if self._user:
            user_info[recordsm.ExtraKeys.USER] = self._user
        if self._hostname:
            user_info[recordsm.ExtraKeys.HOSTNAME] = self._hostname
        return user_info

    def merge(
        self,
        result_set: frontend.ResultSet[object],
        *,
        meta=None,  # pylint: disable=unused-argument
        batch_size=1024,
        progress_callback: Callable[[utils.Progress, Optional[result_types.MergeResult]],
                                    None] = None
    ) -> result_types.MergeResult:
        """Merge a set of objects into this database.

        Given a set of results from another archive this will attempt to merge the corresponding records
        into this historian's archive.

        :param result_set: the set of records to merge from the source historian
        :param meta: option for merging metadata, allowed values:
            None        - Don't merge metadata
            'update'    - Perform dictionary update with existing metadata
            'overwrite' - In the case of an existing metadata dictionary, overwrite it
        """
        # REMOTE
        remote = result_set.historian  # type: Historian
        # Get information about the records that we've been asked to merge
        remote_partial_records = result_set._project(recordsm.OBJ_ID, recordsm.VERSION)  # pylint: disable=protected-access
        remote_snapshot_ids = set(map(recordsm.SnapshotId.from_dict,
                                      remote_partial_records))  # DB HIT

        progress = utils.Progress(len(remote_snapshot_ids))
        if progress_callback is not None:
            progress_callback(progress, result_types.MergeResult())

        result = result_types.MergeResult()
        # get the outgoing snapshot ref. graph
        while remote_snapshot_ids:
            # Get a batch
            batch = set()
            try:
                for _ in range(batch_size):
                    batch.add(remote_snapshot_ids.pop())
            except KeyError:
                pass
            graph = remote.references.get_snapshot_ref_graph(*batch)

            # The graph may contain nodes that are still in our list of remote snapshots to transfer
            # so check and remove these because they will be done in this batch
            extras = set(graph.nodes) - batch
            remote_snapshot_ids.difference_update(extras)

            partial_result = self._merge_batch(remote, graph)
            result.update(partial_result)

            progress.done = progress.total - len(remote_snapshot_ids)
            if progress_callback is not None:
                progress_callback(progress, partial_result)

        return result

    def _merge_batch(self, remote: 'Historian',
                     remote_ref_graph: networkx.DiGraph) -> result_types.MergeResult:
        sid_strings = list(map(str, remote_ref_graph.nodes))

        # REMOTE
        # Get the partial records for all these snapshots indexed by the SID
        remote_partial_records = {}
        for entry in remote.archive.snapshots.find({'_id': qops.in_(*sid_strings)},
                                                   projection={
                                                       recordsm.OBJ_ID: 1,
                                                       recordsm.VERSION: 1,
                                                       recordsm.SNAPSHOT_HASH: 1
                                                   }):  # DB HIT
            remote_partial_records[recordsm.SnapshotId.from_dict(entry)] = entry

        # LOCAL
        # Find the local snapshots along with their hashes
        local_partial_records = {}
        for entry in self.archive.snapshots.find({'_id': qops.in_(*sid_strings)},
                                                 projection={
                                                     recordsm.OBJ_ID: 1,
                                                     recordsm.VERSION: 1,
                                                     recordsm.SNAPSHOT_HASH: 1
                                                 }):  # DB HIT
            local_partial_records[recordsm.SnapshotId.from_dict(entry)] = entry

        # Remove all those that match and log any that have conflicting hashes
        conflicting = []
        for sid, local_partial in local_partial_records.items():
            remote_record = remote_partial_records.pop(sid)
            if remote_record[recordsm.SNAPSHOT_HASH] != local_partial[recordsm.SNAPSHOT_HASH]:
                conflicting.append(sid)

        if conflicting:
            raise exceptions.MergeError(
                'Cannot merge, the following snapshots have conflicting hashes: {}'.format(
                    conflicting))

        # Finally, get all the records to merge and create merge operations
        ops = []
        files_to_transfer = []
        for remote_record in remote.archive.snapshots.find(
            {'_id': qops.in_(*map(str, remote_partial_records.keys()))}):  # DB HIT
            record = recordsm.DataRecord(**remote_record)
            ops.append(operations.Merge(record))
            files_in_record = record.get_files()
            if files_in_record:
                # Extract the second entry in the tuple as this contains the actual state dictionary
                files_to_transfer.extend(entry[1] for entry in files_in_record)

        # and write the new records into our archive
        if ops:
            # Copy the files first.  This way if the user cancels prematurely the files are there but no the objects
            # that refer to them.  The other way around would result in the objects being there but failing when
            # someone tries to load the files
            file_store = self.archive.file_store
            for file_dict in files_to_transfer:
                file_id = file_dict[expr.field_name(files.File.file_id)]
                filename = file_dict[expr.field_name(files.File.filename)] or ''

                with remote.archive.file_store.open_download_stream(file_id) as down_stream:
                    file_store.upload_from_stream_with_id(file_id, filename, down_stream)

            self._archive.bulk_write(ops)  # DB HIT

        return result_types.MergeResult(all_snapshots=remote_ref_graph.nodes,
                                        merged_snapshots=remote_partial_records.keys())

    @contextlib.contextmanager
    def in_transaction(self):
        """This context will either re-use an existing transaction, if one is currently taking place
        or create a new one if not."""
        current = self.current_transaction()
        if current is None:
            ctx = self.transaction()
        else:
            ctx = nullcontext(current)

        with ctx as trans:
            yield trans

    @contextlib.contextmanager
    def transaction(self):
        """Start a new transaction.  Will be nested if there is already one underway"""
        if self._transactions:
            # Start a nested one
            with self._transactions[-1].nested() as nested:
                self._transactions.append(nested)
                try:
                    yield nested
                except Exception:  # Need this so we can have 'else' pylint: disable=try-except-raise
                    raise
                else:
                    self._closing_transaction(nested)
                finally:
                    popped = self._transactions.pop()
                    assert popped is nested
        else:
            # New transaction
            trans = Transaction()
            self._transactions = [trans]

            try:
                yield trans
            except RollbackTransaction:
                pass
            else:
                self._closing_transaction(trans)
                self._commit_transaction(trans)
            finally:
                assert len(self._transactions) == 1
                assert self._transactions[0] is trans
                self._transactions = None

    def current_transaction(self) -> Optional[Transaction]:
        """Get the current transaction if there is one, otherwise returns None"""
        if not self._transactions:
            return None
        return self._transactions[-1]

    def _closing_transaction(self, trans: Transaction):
        conflicting = set()

        # Filter out the deleted records
        del_ops = filter(
            lambda _: isinstance(_, operations.Insert) and _.record.is_deleted_record(),
            trans.staged)

        obj_ids = set(operation.obj_id for operation in del_ops)
        ref_graph = self.references.get_obj_ref_graph(*obj_ids, direction=archives.INCOMING)
        for obj_id in obj_ids:
            for edge in ref_graph.in_edges(obj_id):
                conflicting.add(edge[1])

        if conflicting:
            raise exceptions.ReferenceError('Cannot perform delete', conflicting)

    def _commit_transaction(self, trans: Transaction):
        """Commit a transaction that is finishing"""
        # Perform the database operations first because if these fail we shouldn't update ourselves
        # Save any records that were staged for archiving
        if trans.staged:
            self._archive.bulk_write(trans.staged)

        # Now all is good we can update

        # Live objects
        self._live_objects.update(trans.live_objects)
        # Deleted objects
        for deleted in trans.deleted:
            try:
                self._live_objects.remove(deleted)
            except exceptions.NotFound:
                pass

        # Snapshots
        for ref, obj in trans.snapshots.items():
            self._snapshots_objects[obj] = ref

        # Finally update the metadata as this is least important
        # Metas
        if trans.metas:
            self._archive.meta_set_many(trans.metas)

    def _load_object_from_record(self,
                                 record: recordsm.DataRecord,
                                 depositor: depositors.LiveDepositor = None):

        depositor = depositor or self._live_depositor

        # Try getting the object from the our dict of up to date ones
        obj_id = record.obj_id
        try:
            return self.get_obj(obj_id)
        except exceptions.NotFound:
            pass

        with self.in_transaction() as trans:

            if trans.is_deleted(obj_id):
                raise exceptions.ObjectDeleted(obj_id)

            if record.is_deleted_record():
                raise exceptions.ObjectDeleted(obj_id)

            logger.debug('Loading object from record: %s', record.snapshot_id)
            # Ok, just use the one from the archive
            return depositor.load_from_record(record)

    def _load_object(self, obj_id, depositor: depositors.LiveDepositor):
        obj_id = self._ensure_obj_id(obj_id)

        # Try getting the object from the our dict of up to date ones
        try:
            return self.get_obj(obj_id)
        except exceptions.NotFound:
            pass

        with self.in_transaction() as trans:
            if trans.is_deleted(obj_id):
                raise exceptions.ObjectDeleted(obj_id)

            record = self._objects.records.get(obj_id)

            if record.is_deleted_record():
                raise exceptions.ObjectDeleted(obj_id)

            try:
                obj = self._live_objects.get_object(obj_id)
            except exceptions.NotFound:
                logger.debug('Loading object from record: %s', record.snapshot_id)
                # Ok, just use the one from the archive
                return depositor.load_from_record(record)
            else:
                if record.version != self.get_snapshot_id(obj).version:
                    # The one in the archive is newer, so use that
                    logger.debug('Updating object from record: %s', record.snapshot_id)
                    depositor.update_from_record(obj, record)

                return obj

    def _save_object(self, obj, depositor) -> recordsm.DataRecord:
        with self.in_transaction() as trans:
            try:
                helper = self._ensure_compatible(type(obj))
            except TypeError:
                raise TypeError('Type is incompatible with the historian: {}'.format(
                    type(obj).__name__)) from None

            # Check if an object is already being saved in the transaction
            try:
                record = trans.get_record_for_live_object(obj)
                return record
            except exceptions.NotFound:
                pass

            with self._saving(obj):
                # Ok, have to save it
                current_hash = self.hash(obj)

                try:
                    # Let's see if we have a record at all
                    record = self._live_objects.get_record(obj)
                except exceptions.NotFound:
                    # Object being saved for the first time
                    builder = self._create_builder(obj, snapshot_hash=current_hash)
                    record = depositor.save_from_builder(obj, builder)
                    if self.meta.sticky:
                        # Apply the sticky meta
                        self.meta.update(record.obj_id, self.meta.sticky)
                    return record
                else:
                    if helper.IMMUTABLE:
                        logger.info("Tried to save immutable object with id '%s' again",
                                    record.obj_id)
                        return record

                    # Check if our record is up to date
                    with self.transaction() as nested:
                        loaded_obj = self._new_snapshot_depositor().load_from_record(record)

                        if current_hash == record.snapshot_hash and self.eq(obj, loaded_obj):
                            # Objects identical
                            nested.rollback()
                        else:
                            builder = record.child_builder(snapshot_hash=current_hash)
                            record = depositor.save_from_builder(obj, builder)

                    return record

    def _create_builder(self, obj, **additional) -> recordsm.DataRecordBuilder:
        """Create a record builder for a new object object"""
        additional = additional or {}
        helper = self._ensure_compatible(type(obj))

        builder = recordsm.DataRecord.new_builder(type_id=helper.TYPE_ID,
                                                  obj_id=self._archive.create_archive_id(),
                                                  version=0)
        builder.update(additional)
        return builder

    def _ensure_obj_id(self, obj_or_identifier):
        """
        This call will try and get an object id from the passed parameter.  Uses .to_obj_id() and raises NotFound if it
        is not possible to get the object id.
        """
        obj_id = self.to_obj_id(obj_or_identifier)
        if obj_id is None:
            raise exceptions.NotFound(
                "Could not get an object id from '{}'".format(obj_or_identifier))

        return obj_id

    def _ensure_compatible(self, obj_type: Type[types.SavableObject]) -> helpers.TypeHelper:
        if obj_type not in self._type_registry:
            self.register_type(obj_type)

        return self._type_registry.get_helper_from_obj_type(obj_type)

    def _prepare_obj_id(self, obj_id):
        if obj_id is None:
            return None

        # Convert object ids to the expected type before passing to archive
        try:
            return self._ensure_obj_id(obj_id)
        except exceptions.NotFound as exc:
            # Maybe it is multiple object ids
            if not isinstance(obj_id, Iterable):  # pylint: disable=isinstance-second-argument-not-valid-type
                raise TypeError("Cannot get object id(s) from '{}'".format(obj_id)) from exc

            return list(map(self._ensure_obj_id, obj_id))

    def _prepare_type_id(self, obj_type):
        if obj_type is None:
            return None

        try:
            return self.get_obj_type_id(obj_type)
        except TypeError as exc:
            # Maybe it is multiple type ids
            if not isinstance(obj_type, Iterable):  # pylint: disable=isinstance-second-argument-not-valid-type
                raise TypeError("Cannot get type id(s) from '{}'".format(obj_type)) from exc

            return list(map(self.get_obj_type_id, obj_type))

    def _create_loadable_record(self, record: recordsm.DataRecord) -> 'LoadableRecord':
        return LoadableRecord(record, self.load_snapshot_from_record, self._load_object_from_record)

    @contextlib.contextmanager
    def _saving(self, obj):
        obj_id = id(obj)
        if obj_id in self._saving_set:
            raise RuntimeError(
                'The object is already being saved, this cannot be called twice and suggests '
                'a circular reference is being made')
        self._saving_set.add(obj_id)
        try:
            yield
        finally:
            self._saving_set.remove(obj_id)

    def _record_builder_created(self,
                                builder: recordsm.DataRecordBuilder) -> recordsm.DataRecordBuilder:
        """Update a data record builder with standard information."""
        builder.extras.update(self.get_user_info())
        return builder

    def _new_snapshot_depositor(self):
        return depositors.SnapshotLoader(self)


class SnapshotLoadableRecord(recordsm.DataRecord):
    __slots__ = ()

    def __new__(cls, record_dict: dict, snapshot_loader: Callable[[recordsm.DataRecord], object]):
        loadable = super().__new__(cls, **record_dict)
        loadable._snapshot_loader = snapshot_loader
        return loadable

    def load(self) -> object:
        return self._snapshot_loader(self)


class LoadableRecord(recordsm.DataRecord):
    __slots__ = ()
    _obj_loader = None  # type: Optional[Callable[[recordsm.DataRecord], object]]
    _snapshot_loader = None  # type: Optional[Callable[[recordsm.DataRecord], object]]

    def __new__(cls, record_dict: dict, snapshot_loader: Callable[[recordsm.DataRecord], object],
                obj_loader: Callable[[recordsm.DataRecord], object]):
        loadable = super().__new__(cls, **record_dict)
        loadable._obj_loader = obj_loader
        loadable._snapshot_loader = snapshot_loader
        return loadable

    def load_snapshot(self) -> object:
        return self._snapshot_loader(self)  # pylint: disable=not-callable

    def load(self) -> object:
        return self._obj_loader(self)  # pylint: disable=not-callable



# -*- coding: utf-8 -*-
__all__ = ('NotFound', 'ModificationError', 'ObjectDeleted', 'DuplicateKeyError', 'MigrationError',
           'VersionError', 'IntegrityError', 'ReferenceError', 'ConnectionError', 'MergeError')


class ConnectionError(Exception):  # pylint: disable=redefined-builtin
    """Raise when there is an error in connecting to the backend"""


class NotFound(Exception):
    """Raised when something can not be found in the history"""


class ModificationError(Exception):
    """Raised when a modification of the history encountered a problem"""


class ObjectDeleted(NotFound):
    """Raise when the user tries to interact with a deleted object"""


class DuplicateKeyError(Exception):
    """Indicates that a uniqueness constraint was violated"""


class MigrationError(Exception):
    """Indicates that an error occurred during migration"""


class VersionError(Exception):
    """Indicates a version mismatch between the code and the database"""


class IntegrityError(Exception):
    """Indicates an error that occurred because of an operation that would conflict with a database
    constraint"""


class MergeError(Exception):
    """Indicates that an error occurred when trying to merge"""


class ReferenceError(IntegrityError):  # pylint: disable=redefined-builtin
    """Raised when there is an operation that causes a problem with references for example if
    you try to delete an object that is referenced by another this exception will be raised.  The
    objects ids being referenced will be found in .references.
    """

    def __init__(self, msg, references: set):
        super().__init__(msg)
        self.references = references


class UnorderedError(Exception):
    """Raised when an operation is attempted that assumed an underlying ordering but the data is
    unordered"""


class NotOneError(Exception):
    """Raised when a singlar result is expected but there are in fact more"""



# -*- coding: utf-8 -*-
"""Module for all the built in container and other types"""

from abc import ABCMeta
import collections
import collections.abc
import uuid

from . import base_savable
from .files import File, BaseFile
from . import helpers
from . import records
from . import refs
from .utils import sync
from . import types
from . import type_ids

__all__ = ('List', 'LiveList', 'LiveRefList', 'RefList', 'Str', 'Dict', 'RefDict', 'LiveDict',
           'LiveRefDict', 'BaseFile', 'File')


class _UserType(base_savable.SimpleSavable, metaclass=ABCMeta):
    """Mixin for helping user types to be compatible with the historian.
    These typically have a .data member that stores the actual data (list, dict, str, etc)"""
    ATTRS = ('data',)
    data = None  # placeholder
    # This is an optional type for the data member.  See save_instance_state type for information
    # on when it might be useful
    DATA_TYPE = None  # type: type

    def save_instance_state(self, saver):
        # This is a convenient way of storing primitive data types directly as the state
        # rather than having to be a 'data' member of a dictionary.  This makes it much
        # easier to search these types
        if self.DATA_TYPE is not None and \
                issubclass(self.DATA_TYPE, saver.get_historian().primitives):
            self._on_save(saver)  # Call this here are we aren't going to call up the hierarchy
            return self.data

        return super().save_instance_state(saver)

    def load_instance_state(self, saved_state, loader):
        # See save_instance_state
        if self.DATA_TYPE is not None and \
                issubclass(self.DATA_TYPE, loader.get_historian().primitives):
            self._on_load(loader)  # Call this here are we aren't going to call up the hierarchy
            self.data = saved_state
        else:
            super().load_instance_state(saved_state, loader)


class ObjProxy(_UserType):
    """A simple proxy for any object/data type which can also be a primitive"""
    TYPE_ID = uuid.UUID('d43c2db5-1e8c-428f-988f-8b198accde47')

    def __init__(self, data=None):
        super().__init__()
        self.data = data

    def __call__(self):
        return self.data

    def assign(self, value):
        self.data = value


class Str(collections.UserString, _UserType):
    TYPE_ID = uuid.UUID('350f3634-4a6f-4d35-b229-71238ce9727d')
    DATA_TYPE = str

    def __init__(self, seq):
        collections.UserString.__init__(self, seq)
        _UserType.__init__(self)


class Reffer:
    """Mixin for types that want to be able to create references non-primitive types"""

    # pylint: disable=too-few-public-methods, no-self-use

    def _ref(self, obj):
        """Create a reference for a non-primitive, otherwise use the value"""
        if types.is_primitive(obj):
            return obj

        return self._create_ref(obj)

    def _deref(self, obj):
        """Dereference a non-primitive, otherwise return the value"""
        if types.is_primitive(obj):
            return obj

        return obj()

    def _create_ref(self, obj):
        """Create a reference for a given object"""
        return refs.ObjRef(obj)


# region lists


class List(collections.UserList, _UserType):
    TYPE_ID = uuid.UUID('2b033f70-168f-4412-99ea-d1f131e3a25a')
    DATA_TYPE = list

    def __init__(self, initlist=None):
        collections.UserList.__init__(self, initlist)
        _UserType.__init__(self)


class RefList(collections.abc.MutableSequence, Reffer, _UserType):
    """A list that stores all entries as references in the database except primitives"""
    TYPE_ID = uuid.UUID('091efff5-136d-4ac2-bd59-28f50f151263')
    DATA_TYPE = list

    def __init__(self, init_list=None):
        super().__init__()
        init_list = init_list or []
        self.data = self.DATA_TYPE(self._ref(item) for item in init_list)

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return repr(self.data)

    def __getitem__(self, item):
        return self._deref(self.data[item])

    def __setitem__(self, key, value):
        self.data[key] = self._ref(value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __len__(self):
        return self.data.__len__()

    def insert(self, index, value):
        self.data.insert(index, self._ref(value))


class LiveList(collections.abc.MutableSequence, _UserType):
    """A list that is always in sync with the database"""
    TYPE_ID = uuid.UUID('c83e6206-cd29-4fda-bf76-11fce1681cd9')

    def __init__(self, init_list=None):
        super().__init__()
        init_list = init_list or []
        self.data = RefList(self._create_proxy(item) for item in init_list)

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        if self.is_saved():
            proxy.sync()
        return proxy()

    @sync(save=True)
    def __setitem__(self, key, value):
        proxy = self.data[key]  # type: ObjProxy
        proxy.assign(value)

    @sync(save=True)
    def __delitem__(self, key):
        proxy = self.data[key]  # type: ObjProxy
        del self.data[key]
        if self._historian is not None:
            self._historian.delete(proxy)

    @sync()
    def __len__(self):
        return len(self.data)

    @sync(save=True)
    def insert(self, index, value):
        proxy = self._create_proxy(value)
        self.data.insert(index, proxy)

    @sync(save=True)
    def append(self, value):
        proxy = self._create_proxy(value)
        self.data.append(proxy)

    def _create_proxy(self, obj):  # pylint: disable=no-self-use
        return ObjProxy(obj)


class LiveRefList(Reffer, LiveList):
    """A live list that uses references to store objects"""
    TYPE_ID = uuid.UUID('98454806-c587-4fcc-a514-65fdefb0180d')

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        if self.is_saved():
            proxy.sync()
        ref = proxy()
        return self._deref(ref)

    @sync(save=True)
    def __setitem__(self, key, value):
        proxy = self.data[key]  # type: ObjProxy
        proxy.assign(self._ref(value))

    def _create_proxy(self, obj):
        return ObjProxy(self._ref(obj))


# endregion

# region Dicts


class Dict(collections.UserDict, _UserType):
    TYPE_ID = uuid.UUID('a7584078-95b6-4e00-bb8a-b077852ca510')
    DATA_TYPE = dict

    def __init__(self, *args, **kwarg):
        collections.UserDict.__init__(self, *args, **kwarg)
        _UserType.__init__(self)


class RefDict(collections.abc.MutableMapping, Reffer, _UserType):
    """A dictionary that stores all values as references in the database."""
    TYPE_ID = uuid.UUID('c95f4c4e-766b-4dda-a43c-5fca4fd7bdd0')
    DATA_TYPE = dict

    def __init__(self, *args, **kwargs):
        super().__init__()
        initial = dict(*args, **kwargs)
        self.data = self.DATA_TYPE({key: self._ref(value) for key, value in initial.items()})

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return repr(self.data)

    def __getitem__(self, item):
        return self._deref(self.data[item])

    def __setitem__(self, key, value):
        self.data[key] = self._ref(value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return self.data.__len__()


class LiveDict(collections.abc.MutableMapping, _UserType):
    TYPE_ID = uuid.UUID('740cc832-721c-4f85-9628-706257eb55b9')
    DATA_TYPE = RefDict

    def __init__(self, *args, **kwargs):
        super().__init__()
        initial = dict(*args, **kwargs)
        self.data = RefDict({key: self._create_proxy(value) for key, value in initial.items()})

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        proxy.sync()
        return proxy()

    @sync()
    def __setitem__(self, key, value):
        if key in self.data:
            # Can simply update the proxy
            proxy = self.data[key]  # type: ObjProxy
            proxy.assign(value)
            if proxy.is_saved():
                proxy.save()
        else:
            proxy = self._create_proxy(value)
            self.data[key] = proxy
            if self.is_saved():
                self.save()  # This will cause the proxy to be saved as well

    @sync(save=True)
    def __delitem__(self, key):
        proxy = self.data[key]
        del self.data[key]
        self._historian.delete(proxy)

    @sync()
    def __iter__(self):
        return self.data.__iter__()

    @sync()
    def __len__(self):
        return len(self.data)

    def _create_proxy(self, value):  # pylint: disable=no-self-use
        return ObjProxy(value)


class LiveRefDict(Reffer, LiveDict):
    """A live dictionary that uses references to refer to contained objects"""
    TYPE_ID = uuid.UUID('16e7e814-8268-46e0-8d8e-6f34132366b9')

    def __init__(self, *args, **kwargs):
        super().__init__()

        initial = dict(*args, **kwargs)
        self.data = RefDict({key: self._create_proxy(value) for key, value in initial.items()})

    @sync()
    def __getitem__(self, item):
        proxy = self.data[item]  # type: ObjProxy
        proxy.sync()
        ref = proxy()
        return self._deref(ref)

    @sync()
    def __setitem__(self, key, value):
        if key in self.data:
            # Can simply update the proxy
            proxy = self.data[key]  # type: ObjProxy
            proxy.assign(self._ref(value))
            if proxy.is_saved():
                proxy.save()
        else:
            proxy = self._create_proxy(value)
            self.data[key] = proxy
            if self.is_saved():
                self.save()  # This will cause the proxy to be saved as well

    def _create_proxy(self, value):
        return ObjProxy(self._ref(value))


# endregion


class SnapshotIdHelper(helpers.TypeHelper):
    """Add ability to store references"""
    TYPE = records.SnapshotId
    TYPE_ID = type_ids.SNAPSHOT_ID_TYPE_ID

    def eq(self, one, other):  # pylint: disable=invalid-name, no-self-use
        if not (isinstance(one, records.SnapshotId) and isinstance(other, records.SnapshotId)):
            return False

        return one.obj_id == other.obj_id and one.version == other.version

    def yield_hashables(self, obj, hasher):  # pylint: disable=no-self-use
        yield from hasher.yield_hashables(obj.obj_id)
        yield from hasher.yield_hashables(obj.version)

    def save_instance_state(self, obj, _saver):  # pylint: disable=no-self-use
        return obj.to_dict()

    def load_instance_state(self, obj, saved_state, _loader):  # pylint: disable=no-self-use
        if isinstance(saved_state, list):
            # Legacy version
            obj.__init__(*saved_state)
        else:
            # New version is a dictionary
            obj.__init__(**saved_state)


HISTORIAN_TYPES = (Str, List, RefList, LiveList, LiveRefList, Dict, RefDict, LiveDict, LiveRefDict,
                   ObjProxy, File)



# -*- coding: utf-8 -*-
"""Classes and function useful for trying out mincepy functionality"""
# pylint: disable=cyclic-import
import contextlib
import logging
import os
import string
import random
from typing import Iterator
import uuid

import bson
import pymongo

import mincepy

logger = logging.getLogger(__name__)

ENV_ARCHIVE_URI = 'MINCEPY_TEST_URI'
ENV_ARCHIVE_BASE_URI = 'MINCEPY_TEST_BASE_URI'
DEFAULT_ARCHIVE_URI = 'mongodb://localhost/mincepy-tests'
DEFAULT_ARCHIVE_BASE_URI = 'mongodb://127.0.0.1'

# pylint: disable=redefined-outer-name, invalid-name


def get_base_uri() -> str:
    """Get a base URI for an archive that can be used for testing.  This will not contain the database name as multiple
    databases can be used during a test session."""
    return os.environ.get(ENV_ARCHIVE_BASE_URI, DEFAULT_ARCHIVE_BASE_URI)


def create_archive_uri(base_uri='', db_name=''):
    """Get an archive URI based on the current archive base URI plus the passed database name.

    If the database name is missing a random one will be used"""
    if not db_name:
        letters = string.ascii_lowercase
        db_name = 'mincepy-' + ''.join(random.choice(letters) for _ in range(5))
    base_uri = base_uri or get_base_uri()
    return base_uri + '/' + db_name


@contextlib.contextmanager
def temporary_archive(archive_uri: str) -> Iterator[mincepy.Archive]:
    """Create a temporary archive.  The associated database will be dropped on exiting the context"""
    client = pymongo.MongoClient(archive_uri)
    db = client.get_default_database()  # pylint: disable=invalid-name
    client.drop_database(db)
    mongo_archive = mincepy.mongo.MongoArchive(db)
    yield mongo_archive
    client.drop_database(db)


@contextlib.contextmanager
def temporary_historian(archive_uri: str = '') -> Iterator[mincepy.Archive]:
    """Create a temporary historian.  The associated database will be dropped on exiting the context."""
    with temporary_archive(archive_uri) as archive:
        yield mincepy.Historian(archive)


try:
    import pytest

    # Optional pytest fixtures

    @pytest.fixture
    def archive_uri() -> str:
        return os.environ.get(ENV_ARCHIVE_URI, DEFAULT_ARCHIVE_URI)

    @pytest.fixture
    def archive_base_uri() -> str:
        return os.environ.get(ENV_ARCHIVE_BASE_URI, DEFAULT_ARCHIVE_BASE_URI)

    @pytest.fixture
    def mongodb_archive(archive_uri):
        with temporary_archive(archive_uri) as mongo_archive:
            yield mongo_archive

    @pytest.fixture(autouse=True)
    def historian(mongodb_archive):
        hist = mincepy.Historian(mongodb_archive)
        hist.register_types(mincepy.testing.HISTORIAN_TYPES)
        mincepy.set_historian(hist)
        yield hist
        mincepy.set_historian(None)

except ImportError:
    logger.debug("pytest fixtures missing because pytest isn't installed")


class Car(mincepy.ConvenientSavable):
    TYPE_ID = bson.ObjectId('5e075d6244572f823ed93274')
    colour = mincepy.field()
    make = mincepy.field()

    def __init__(self, make='ferrari', colour='red'):
        super().__init__()
        self.make = make
        self.colour = colour

    def __str__(self) -> str:
        return '{} {}'.format(self.colour, self.make)


class Garage(mincepy.ConvenientSavable):
    TYPE_ID = bson.ObjectId('5e07b40a44572f823ed9327b')
    car = mincepy.field()

    def __init__(self, car=None):
        super().__init__()
        self.car = car

    def save_instance_state(self, saver: mincepy.Saver):
        state = super().save_instance_state(saver)
        return state


class Person(mincepy.ConvenientSavable):
    TYPE_ID = uuid.UUID('d60ca740-9fa6-4002-83f6-e4c91403e41b')
    name = mincepy.field()
    age = mincepy.field()
    car = mincepy.field(ref=True)

    def __init__(self, name, age, car=None):
        super().__init__()
        self.name = name
        self.age = age
        self.car = car


class Cycle(mincepy.ConvenientSavable):
    TYPE_ID = uuid.UUID('600fb6ae-684c-4f8e-bed3-47ae06739d29')
    _ref = mincepy.field()

    def __init__(self, ref=None):
        super().__init__()
        self._ref = mincepy.ObjRef(ref)

    @property
    def ref(self):
        return self._ref()

    @ref.setter
    def ref(self, value):
        self._ref = mincepy.ObjRef(value)

    def __eq__(self, other):
        return self.ref is other.ref

    def yield_hashables(self, hasher):
        yield from hasher.yield_hashables(id(self.ref))


def populate(historian=None):
    historian = historian or mincepy.get_historian()

    colours = ('red', 'orange', 'yellow', 'green', 'blue', 'indigo', 'violet')
    makes = ('honda', 'ferrari', 'zonda', 'fiat')

    cars = []

    for make in makes:
        for colour in colours:
            # Make some cars
            car = Car(make, colour)
            historian.save(car)
            cars.append(car)

    # Now randomly change some of them
    for _ in range(int(len(cars) / 4)):
        car = random.choice(cars)
        car.colour = random.choice(colours)
        car.save()

    # Now change one a number of times
    car = random.choice(cars)
    for colour in colours:
        car.colour = colour
        car.save()

    people = mincepy.RefList()
    for name in ('martin', 'sonia', 'gavin', 'upul', 'martin', 'sebastiaan', 'irene'):
        person = Person(name, random.randint(20, 40))
        historian.save(person)
        people.append(person)
    historian.save(people)


HISTORIAN_TYPES = Car, Garage, Person, Cycle



# -*- coding: utf-8 -*-
"""Query expressions"""
import abc
import copy
from typing import Union, List, Iterable

__all__ = ('Expr', 'WithListOperand', 'Empty', 'Operator', 'Eq', 'Gt', 'Gte', 'In', 'Lt', 'Lte',
           'Ne', 'Nin', 'Comparison', 'Logical', 'And', 'Not', 'Or', 'Nor', 'Exists', 'Queryable',
           'WithQueryContext', 'query_expr', 'field_name', 'build_expr', 'Query')


class FilterLike(metaclass=abc.ABCMeta):
    """An abstract base class for objects representing a pyos path, e.g. pyos.pathlib.PurePath."""

    # pylint: disable=too-few-public-methods

    @abc.abstractmethod
    def __query_expr__(self) -> dict:
        """Return the query filter representation of the object."""


FilterSpec = Union[dict, FilterLike]


class Expr(FilterLike, metaclass=abc.ABCMeta):
    """The base class for query expressions.  Expressions are tuples containing an operator or a
    field as a first part and a value or expression as second"""
    __slots__ = ()

    def dict(self):
        """Return the query dictionary for this expression"""
        return self.__query_expr__()

    def __and__(self, other: 'Expr') -> 'And':
        if not isinstance(other, Expr):
            raise TypeError("Expected Expr got '{}'".format(other))
        return And([self, other])

    def __or__(self, other: 'Expr') -> 'Or':
        if not isinstance(other, Expr):
            raise TypeError("Expected Expr got '{}'".format(other))
        return Or([self, other])


class WithListOperand(FilterLike):
    """Mixin for expressions that take an operand that is a list"""

    # pylint: disable=no-member, too-few-public-methods

    def __init__(self, operand: List[Expr]):
        if not isinstance(operand, list):
            raise TypeError('Expected a list, got {}'.format(type(operand).__name__))
        for entry in operand:
            if not isinstance(entry, Expr):
                raise TypeError('Expected a list of Expr, found {}'.format(type(entry).__name__))
        self.operand = operand

    def __query_expr__(self) -> dict:
        if len(self.operand) == 1:
            return query_expr(self.operand[0])

        return {self.oper: list(map(query_expr, self.operand))}


class Empty(Expr):
    """The empty expression"""

    def __query_expr__(self) -> dict:
        return {}


# region Match


class Operator(Expr):
    """An simple operator expression.

    Consists of an operator applied to an operand which is to be matched
    """

    __slots__ = ('value',)
    oper = None  # type: str

    def __init__(self, value):
        self.value = value

    def __query_expr__(self) -> dict:
        return {self.oper: self.value}


class Eq(Operator):
    __slots__ = ()
    oper = '$eq'


class Gt(Operator):
    __slots__ = ()
    oper = '$gt'


class Gte(Operator):
    __slots__ = ()
    oper = '$gte'


class In(Operator):
    __slots__ = ()
    oper = '$in'


class Lt(Operator):
    __slots__ = ()
    oper = '$lt'


class Lte(Operator):
    __slots__ = ()
    oper = '$lte'


class Ne(Operator):
    __slots__ = ()
    oper = '$ne'


class Nin(Operator):
    __slots__ = ()
    oper = '$nin'


COMPARISON_OPERATORS = {oper_type.oper: oper_type for oper_type in Operator.__subclasses__()}


class Comparison(Expr):
    """A comparison expression consists of a field and an operator expression e.g. name == 'frank'
    where name is the field, the operator is ==, and the value is 'frank'
    """
    __slots__ = 'field', 'expr'

    def __init__(self, field, expr: Operator):
        if field is None:
            raise ValueError('field cannot be None')
        if not isinstance(expr, Operator):
            raise TypeError("Expected an operator expression, got '{}'".format(type(expr).__name__))

        self.field = field
        self.expr = expr

    def __query_expr__(self) -> dict:
        if isinstance(self.expr, Eq):
            # Special case for this query as it looks nicer this way (without using '$eq')
            return {field_name(self.field): self.expr.value}

        return {field_name(self.field): query_expr(self.expr)}


# endregion

# region Logical operators


class Logical(Expr):
    """A comparison operation.  Consists of an operator applied to an operand which is matched in a
    particular way"""

    __slots__ = ('operand',)
    oper = None  # type: str

    def __init__(self, operand: Expr):
        if not isinstance(operand, Expr):
            raise TypeError("Expected an Expr, got '{}'".format(type(operand).__name__))
        self.operand = operand

    def __query_expr__(self) -> dict:
        return {self.oper: query_expr(self.operand)}


class And(WithListOperand, Logical):
    __slots__ = ()
    oper = '$and'

    def __and__(self, other: 'Expr') -> 'And':
        if isinstance(other, And):
            # Economise on Ands and fuse them here
            return And([*self.operand, *other.operand])

        return super().__and__(other)


class Not(Logical):
    __slots__ = ()
    oper = '$not'


class Or(WithListOperand, Logical):
    __slots__ = ()
    oper = '$or'

    def __or__(self, other: 'Expr') -> 'Or':
        if isinstance(other, Or):
            # Economise on Ors and fuse them here
            return Or([*self.operand, *other.operand])

        return super().__or__(other)


class Nor(WithListOperand, Logical):
    __slots__ = ()
    oper = '$nor'


# endregion

# region Element operators


class Exists(Operator):
    __slots__ = ()
    oper = '$exists'


# endregion


class Queryable(metaclass=abc.ABCMeta):
    # region Query operations
    __slots__ = ()
    __hash__ = object.__hash__

    def __eq__(self, other) -> Comparison:
        return Comparison(self.get_path(), Eq(other))

    def __ne__(self, other) -> Comparison:
        return Comparison(self.get_path(), Ne(other))

    def __gt__(self, other) -> Comparison:
        return Comparison(self.get_path(), Gt(other))

    def __ge__(self, other) -> Comparison:
        return Comparison(self.get_path(), Gte(other))

    def __lt__(self, other) -> Comparison:
        return Comparison(self.get_path(), Lt(other))

    def __le__(self, other) -> Comparison:
        return Comparison(self.get_path(), Lte(other))

    def in_(self, *possibilities) -> Comparison:
        return Comparison(self.get_path(), In(possibilities))

    def nin_(self, *possibilities) -> Expr:
        return Comparison(self.get_path(), Nin(possibilities))

    def exists_(self, value: bool = True) -> Expr:
        return Comparison(self.get_path(), Exists(value))

    @abc.abstractmethod
    def get_path(self) -> str:
        """Get the path for this object in the document"""


class WithQueryContext:
    """A mixin for Queryable objects that allows a context to be added which is always 'anded' with
    the resulting query condition for any operator"""
    _query_context = None

    # pylint: disable=no-member

    def set_query_context(self, expr: Expr):
        self._query_context = expr

    def __eq__(self, other) -> Expr:
        return self._combine(super().__eq__(other))

    def __ne__(self, other) -> Expr:
        return self._combine(super().__ne__(other))

    def __gt__(self, other) -> Expr:
        return self._combine(super().__gt__(other))

    def __ge__(self, other) -> Expr:
        return self._combine(super().__ge__(other))

    def __lt__(self, other) -> Expr:
        return self._combine(super().__lt__(other))

    def __le__(self, other) -> Expr:
        return self._combine(super().__le__(other))

    def in_(self, *possibilities) -> Expr:
        return self._combine(super().in_(*possibilities))

    def nin_(self, *possibilities) -> Expr:
        return self._combine(super().nin_(*possibilities))

    def exists_(self, value: bool = True) -> Expr:
        return self._combine(super().exists_(value))

    def _combine(self, expr: Expr) -> Expr:
        if self._query_context is None:
            return expr
        return And([self._query_context, expr])


def query_expr(filter: FilterLike) -> dict:  # pylint: disable=redefined-builtin
    """Return a query specification (dict)

    If a dict is passed is is returned unaltered.
    Otherwise __qspec__() is called and its value is returned as long as it is a dict. In all other
    cases, TypeError is raised."""
    if isinstance(filter, dict):
        return filter

    # Work from the object's type to match method resolution of other magic methods.
    try:
        query_repr = filter.__query_expr__()
    except AttributeError:
        raise TypeError('expected dict or object with __query_expr__, not ' + str(filter)) from None

    if isinstance(query_repr, dict):
        return query_repr

    raise TypeError('expected {}.__query_expr__() to return dict, not {}'.format(
        type(filter).__name__,
        type(query_repr).__name__))


def field_name(field) -> str:
    if isinstance(field, str):
        return field

    try:
        name = field.__field_name__()
    except AttributeError:
        raise TypeError('expected str or object with __field__name__, not ' + str(field)) from None

    if isinstance(name, str):
        return name

    raise TypeError('expected {}.__field_name__() to return str, not {}'.format(
        type(field).__name__,
        type(name).__name__))


def build_expr(item) -> Expr:
    """Expression factory"""
    # pylint: disable=too-many-branches, too-many-return-statements

    if isinstance(item, Expr):
        return item

    if isinstance(item, dict):
        if not item:
            return Empty()

        if len(item) == 1:
            return build_expr(tuple(item.items())[0])

        # Otherwise a dictionary is an implicit 'and'
        return And(list(map(build_expr, item.items())))

    if isinstance(item, tuple):
        if len(item) != 2:
            raise ValueError('Expecting tuple of length two, instead got {}'.format(item))

        first, second = item
        if first.startswith('$'):
            # Comparison operators
            try:
                oper = COMPARISON_OPERATORS[first]
            except KeyError:
                pass
            else:
                return oper(second)

            # Logical operators
            if first == '$and':
                return And(list(map(build_expr, second)))
            if first == '$not':
                return Not(build_expr(second))
            if first == '$nor':
                return Nor(list(map(build_expr, second)))
            if first == '$or':
                return Or(list(map(build_expr, second)))
            # Element query
            if first == '$exists':
                return Exists(second)

            raise ValueError("Unknown operator '{}'".format(item))

        # Must be a 'match' where the first is the field
        try:
            return Comparison(first, build_expr(second))
        except (ValueError, TypeError):
            # pylint: disable=fixme
            # TODO: See if we can make this check safer
            # Assume second is a value type
            return Comparison(first, Eq(second))

    try:
        return item.__expr__()
    except AttributeError:
        raise TypeError('expected dict or object with __expr__, not ' +
                        type(item).__name__) from None


class Query:
    __slots__ = '_filter_expressions', 'limit', 'sort', 'skip'

    def __init__(self, *expr: Expr, limit: int = None, sort: dict = None, skip: int = None):
        self._filter_expressions = []  # type: List[Expr]
        self.extend(expr)
        self.limit = limit
        self.sort = sort
        self.skip = skip

    def __str__(self) -> str:
        return str(self.__dict__)

    def copy(self) -> 'Query':
        return Query(*copy.copy(self._filter_expressions),
                     limit=self.limit,
                     sort=self.sort,
                     skip=self.skip)

    @property
    def __dict__(self) -> dict:
        return dict(filter=self.get_filter(), sort=self.sort, limit=self.limit, skip=self.skip)

    def append(self, expr: Expr):
        self._filter_expressions.append(build_expr(expr))

    def extend(self, exprs: Iterable[Expr]):
        for entry in exprs:
            self.append(entry)

    def get_filter(self) -> dict:
        """Get the query filter as a dictionary"""
        if not self._filter_expressions:
            return {}
        return query_expr(And(self._filter_expressions))



# -*- coding: utf-8 -*-
"""Module that contains methods and classes for dealing with database storable attributes of
objects"""

import abc
from typing import Dict, Type

from . import expr

__all__ = ('field',)

_UNSET = ()


class FieldProperties:
    """Properties of a mincePy field"""

    __slots__ = ('store_as', 'attr_name', 'ref', 'dynamic', 'field_type', 'default', 'extras',
                 'db_class')

    # pylint: disable=too-many-arguments
    def __init__(self,
                 attr: str = None,
                 store_as: str = None,
                 ref=False,
                 dynamic=True,
                 field_type: Type = None,
                 default=_UNSET,
                 extras=None):
        """
        Fixed properties of a field

        :param store_as: the name to use for this field when storing in the database
        :param attr: the name of the class attribute
        """
        if store_as and '.' in store_as:
            raise ValueError("store_as cannot contain a dot, got '{}'".format(store_as))

        self.attr_name = attr
        self.store_as = store_as if store_as else attr
        self.ref = ref
        self.dynamic = dynamic
        self.field_type = field_type
        self.default = default
        self.extras = extras or {}
        self.db_class = None

    def __repr__(self):
        return 'FieldProperties(attr={}, store_as={}, ref={}, dynamic={}, field_type={}, ' \
               'default={}, extras={})'.format(
            repr(self.attr_name),
            repr(self.store_as),
            self.ref,
            self.dynamic,
            repr(self.field_type),
            repr(self.default),
            repr(self.extras)
        )

    def class_created(self, the_class: type, attr: str):
        """Called by the metaclass when the owning class is created, should only be done once"""
        assert self.db_class is None, 'Cannot call class_created more than once'
        self.db_class = the_class
        # Don't overwrite these two, they have been set manually and should be respected
        if self.attr_name is None:
            self.attr_name = attr
        if self.store_as is None:
            self.store_as = attr


class Field(expr.WithQueryContext, expr.Queryable, property):
    """Database field class.  Provides information about how to store object attributes in the
    database"""
    __doc__ = ''

    def __init__(self, properties: FieldProperties, path_prefix=''):
        """
        Create a attribute field that will be stored in the database
        """
        super().__init__()
        self._properties = properties
        self.path_prefix = path_prefix

    def __getattribute__(self, item: str):
        try:
            return object.__getattribute__(self, item)
        except AttributeError as exc:
            # Dynamically create a new field
            if item != '__isabstractmethod__':
                if self._properties.field_type is not None and \
                        issubclass(self._properties.field_type, WithFields):
                    properties = get_field_properties(self._properties.field_type)
                    try:
                        child_field = type(self)(properties[item], path_prefix=self.get_path())
                    except KeyError:
                        raise exc from None
                    else:
                        child_field.set_query_context(self._query_context)
                        return child_field

                if self._properties.dynamic:
                    # Creating a dynamic child
                    new_field = type(self)(FieldProperties(store_as=item, attr=item, dynamic=True),
                                           path_prefix=self.get_path())
                    new_field.set_query_context(self._query_context)
                    return new_field

            raise

    def __field_name__(self) -> str:
        return self._properties.store_as

    def __call__(self, fget=None, fset=None, fdel=None, doc=None, prop_kwargs=None):
        """This method allows the field to become a property"""
        self.getter(fget)
        self.setter(fset)
        self.deleter(fdel)
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc
        return self

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return self._getter(obj)

    def __set__(self, obj, value):
        if self._setter is None:
            raise AttributeError("can't set attribute '{}'".format(self._properties.attr_name))
        self._setter(obj, value)

    def __delete__(self, obj):
        if self._deleter is None:
            raise AttributeError("can't delete attribute '{}'".format(self._properties.attr_name))
        self._deleter(obj)

    def getter(self, fget):
        setattr(self, '_getter', fget)
        return self

    def setter(self, fset):
        setattr(self, '_setter', fset)
        return self

    def deleter(self, fdel):
        setattr(self, '_deleter', fdel)
        return self

    def _getter(self, obj):
        """Default getter"""
        try:
            return obj.__dict__[self._properties.attr_name]
        except KeyError:
            raise AttributeError("unreadable attribute '{}'".format(
                self._properties.attr_name)) from None

    def _setter(self, obj, value):
        """Default setter"""
        obj.__dict__[self._properties.attr_name] = value

    def _deleter(self, obj):
        """Default deleter"""
        del obj.__dict__[self._properties.attr_name]

    def get_path(self) -> str:
        if self.path_prefix:
            return self.path_prefix + '.' + self._properties.store_as

        return self._properties.store_as


def field(
        attr: str = None,
        ref=False,
        default=_UNSET,
        type=None,  # pylint: disable=redefined-builtin
        store_as: str = None,
        dynamic=False) -> Field:
    """Define a new field"""
    properties = FieldProperties(attr=attr,
                                 ref=ref,
                                 store_as=store_as,
                                 default=default,
                                 field_type=type,
                                 dynamic=dynamic)
    return Field(properties)


class WithFieldMeta(abc.ABCMeta):
    """Metaclass for database types"""

    def __init__(cls, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)
        for key, value in namespace.items():
            if isinstance(value, Field):
                cls.init_field(value, key)

    # Make this class a mapping such that fields can be accessed using [] operator

    def __iter__(cls):
        return get_fields(cls).__iter__()

    def __len__(cls):
        return get_fields(cls).__len__()

    def __getitem__(cls, item):
        return get_fields(cls).__getitem__(item)


class WithFields(metaclass=WithFieldMeta):
    """Base class for types that describe how to save objects in the database using db fields"""

    @classmethod
    def init_field(cls, obj_field, attr_name: str):
        obj_field._properties.class_created(cls, attr_name)  # pylint: disable=protected-access

    def __init__(self, **kwargs):
        for name, field_properties in get_field_properties(type(self)).items():
            try:
                passed_value = kwargs.pop(name)
            except KeyError:
                # Let's see if there's a default
                if field_properties.default is not _UNSET:
                    setattr(self, field_properties.attr_name, field_properties.default)
            else:
                setattr(self, field_properties.attr_name, passed_value)

        if kwargs:
            raise ValueError("Got unexpected keyword argument(s) '{}'".format(kwargs))


def get_fields(db_type: Type[WithFields]) -> Dict[str, Field]:
    """Given a WithField type this will return all the database attributes as a dictionary where the
    key is the attribute name"""
    db_attrs = {}
    for entry in reversed(db_type.__mro__):
        if entry is object:
            continue
        for name, class_attr in entry.__dict__.items():
            if isinstance(class_attr, Field):
                db_attrs[name] = class_attr

    return db_attrs


def get_field_properties(db_type: Type[WithFields]) -> Dict[str, FieldProperties]:
    """Given a WithField type this will return all the database attributes as a dictionary where the
    key is the attribute name"""
    db_attrs = {}
    for entry in reversed(db_type.__mro__):
        if entry is object:
            continue
        for name, class_attr in entry.__dict__.items():
            if isinstance(class_attr, Field):
                db_attrs[name] = class_attr._properties  # pylint: disable=protected-access

    return db_attrs



import uuid

import mincepy
import mincepy.builtins
from mincepy.testing import Car


def test_save_as_ref(historian: mincepy.Historian):
    """Test the 'AsRef' functionality of BaseSavableObject"""

    class Person(mincepy.SimpleSavable):
        TYPE_ID = uuid.UUID('692429b6-a08b-489a-aa09-6eb3174b6405')
        ATTRS = (mincepy.AsRef('car'), 'name')  # Save the car by reference

        def __init__(self, name: str, car):
            super().__init__()
            self.name = name
            self.car = car

    car = Car()
    # Both martin and sonia have the same car
    martin = Person('martin', car)
    sonia = Person('sonia', car)
    martin_id, sonia_id = historian.save(martin, sonia)
    del martin, sonia, car

    # No reload and check they still have the same car
    martin, sonia = historian.load(martin_id, sonia_id)
    assert martin.car is not None
    assert martin.name == 'martin'
    assert martin.car is sonia.car
    assert sonia.name == 'sonia'



# -*- coding: utf-8 -*-
"""Module for methods related to saving and loading objects to/from records"""

from typing import Type, Union

from . import fields
from . import refs


def save_instance_state(obj, db_type: Type[fields.WithFields] = None):
    """Save the instance state of an object.

    Given an object this function takes a DbType specifying the attributes to be saved and will used
    these to return a saved sate.  Note, that for regular Savable objects, the db_type is the object
    itself in which case this argument can be omitted.
    """
    if db_type is None:
        assert issubclass(type(obj), fields.WithFields), \
            "A DbType wasn't passed and obj isn't a DbType instance other"
        db_type = type(obj)

    field_properties = fields.get_field_properties(db_type).values()
    state = {}

    for properties in field_properties:
        attr_val = getattr(obj, properties.attr_name)
        if properties.ref and attr_val is not None:
            attr_val = refs.ObjRef(attr_val)

        # Check if it's still a field because otherwise it hasn't been set yet
        if attr_val is not properties:
            state[properties.store_as] = attr_val

    return state


def load_instance_state(obj,
                        state: Union[list, dict],
                        db_type: Type[fields.WithFields] = None,
                        ignore_missing=True):
    if db_type is None:
        assert issubclass(type(obj), fields.WithFields), \
            "A DbType wasn't passed and obj isn't a DbType instance other"
        db_type = type(obj)

    to_set = {}
    if isinstance(state, dict):
        for properties in fields.get_field_properties(db_type).values():
            try:
                value = state[properties.store_as]
            except KeyError:
                if ignore_missing:
                    value = None
                else:
                    raise ValueError("Saved state missing '{}'".format(
                        properties.store_as)) from None

            if properties.ref and value is not None:
                assert isinstance(value, refs.ObjRef), \
                    'Expected to see a reference in the saved state for key ' \
                    "'{}' but got '{}'".format(properties.store_as, value)
                value = value()  # Dereference it

            to_set[properties.attr_name] = value

    for attr_name, value in to_set.items():
        setattr(obj, attr_name, value)



# -*- coding: utf-8 -*-
# pylint: disable=unused-import, redefined-outer-name
import random

import pytest

import mincepy
from mincepy import testing
from mincepy.testing import archive_uri, mongodb_archive, historian, archive_base_uri
from . import utils


@pytest.fixture
def standard_dataset(historian: mincepy.Historian):
    with historian.transaction():
        # Put in some cars
        ferrari = testing.Car('ferrari', 'red')
        ferrari.save()
        honda = testing.Car('honda', 'white')
        honda.save()
        fiat = testing.Car('fiat', 'green')
        fiat.save()

        # Put in some identical yellow cars
        for _ in range(4):
            testing.Car(make='renault', colour='yellow').save()

        # Put in some people
        testing.Person('sonia', 30, ferrari).save()
        testing.Person('martin', 35, honda).save()
        testing.Person('gavin', 34).save()
        testing.Person('upul', 35, fiat).save()

        # Put in some people with the same name and different age
        for _ in range(100):
            testing.Person(name='mike', age=random.randint(0, 4)).save()


@pytest.fixture
def large_dataset(historian: mincepy.Historian):
    with historian.transaction():
        # Put in some cars
        ferrari = testing.Car('ferrari', 'red')
        ferrari.save()
        honda = testing.Car('honda', 'white')
        honda.save()
        fiat = testing.Car('fiat', 'green')
        fiat.save()

        # Put in some yellow cars
        for _ in range(100):
            testing.Car(make='renault', colour='yellow').save()

        # Put in some purely random
        for _ in range(100):
            testing.Car(make=utils.random_str(5), colour=utils.random_str(3)).save()

        # Put in some people
        testing.Person('sonia', 30, ferrari).save()
        testing.Person('martin', 35, honda).save()
        testing.Person('gavin', 34).save()
        testing.Person('upul', 35, fiat).save()

        # Put in some with the same age
        for _ in range(100):
            testing.Person(name=utils.random_str(5), age=54).save()

        # Put in some with the same name
        for _ in range(100):
            testing.Person(name='mike', age=random.randint(0, 100)).save()

        # Put in some purely random
        for _ in range(100):
            testing.Person(name=utils.random_str(5), age=random.randint(0, 100)).save()



# -*- coding: utf-8 -*-
import mincepy
from mincepy import frontend
from mincepy import testing

# pylint: disable=invalid-name


def test_collection(historian):

    def identity(x):
        return x

    coll = frontend.EntriesCollection(historian, historian.archive.objects, entry_factory=identity)
    p1 = testing.Person('martin', 35)
    p1.save()
    p2 = testing.Person('john', 5)
    p2.save()

    p1.age = 36
    p1.save()

    # Find by obj ID
    records = coll.find(obj_id=p1.obj_id).one()
    assert isinstance(records, dict)
    assert records['obj_id'] == p1.obj_id

    # Find using multiple obj IDs
    records = list(coll.find(obj_id=[p1.obj_id, p2.obj_id]))
    assert len(records) == 2
    assert {records[0]['obj_id'], records[1]['obj_id']} == {p1.obj_id, p2.obj_id}

    c1 = testing.Car()
    c1.save()

    records = list(coll.find(obj_type=[testing.Car.TYPE_ID, testing.Person.TYPE_ID]))
    assert len(records) == 3
    assert {records[0]['type_id'], records[1]['type_id'], records[2]['type_id']} == \
           {testing.Car.TYPE_ID, testing.Person.TYPE_ID}


def test_distinct(historian):
    """Test that .distinct() works on collections"""
    honda = testing.Car('honda', 'green')
    porsche = testing.Car('porsche', 'black')
    red_honda = testing.Car('honda', 'red')
    fiat = testing.Car('fiat', 'green')
    red_porsche = testing.Car('porsche', 'red')
    historian.save(honda, porsche, red_honda, fiat, red_porsche)

    assert set(historian.objects.distinct('state.colour')) == {'green', 'black', 'red', 'green'}
    # Now using the field
    assert set(historian.objects.distinct(testing.Car.colour)) == {'green', 'black', 'red', 'green'}
    assert len(list(historian.objects.distinct(mincepy.DataRecord.obj_id))) == 5



# -*- coding: utf-8 -*-
"""Module for storing type ids of common mincePy types"""
import uuid

FILE_TYPE_ID = uuid.UUID('3bf3c24e-f6c8-4f70-956f-bdecd7aed091')
OBJ_REF_TYPE_ID = uuid.UUID('633c7035-64fe-4d87-a91e-3b7abd8a6a28')
SNAPSHOT_ID_TYPE_ID = uuid.UUID('05fe092b-07b3-4ffc-8cf2-cee27aa37e81')



# -*- coding: utf-8 -*-
"""
This module collects all the frontend database entities such as collections and results that a
user interacts with (through the historian)
"""
import functools
from typing import TypeVar, Generic, Iterable, Callable, Any, Iterator, Optional, Union

from . import archives
from . import exceptions
from . import expr
from . import fields
from . import records

T = TypeVar('T')  # The type stored by the collection pylint: disable=invalid-name


class ResultSet(Generic[T]):
    """The representation of the results of a query.

    The actual execution of queries is delayed until the user actually requests data through one
    of the methods (including iteration) of this class.

    In general the user should not instantiate this directly but will be returned instances from
    various mincepy methods.
    """

    __slots__ = ('_historian', '_archive_collection', '_query', '_kwargs', '_entry_factory')

    def __init__(self,
                 historian,
                 archive_collection: archives.Collection,
                 query: expr.Query,
                 kwargs: dict = None,
                 entry_factory: Callable[[Any], T] = None):
        self._historian = historian
        self._archive_collection = archive_collection
        self._query = query
        self._kwargs = kwargs or {}
        self._entry_factory = entry_factory or (lambda x: x)

    def __iter__(self) -> Iterable[T]:
        for entry in self._archive_collection.find(**self._query.__dict__, **self._kwargs):
            yield self._entry_factory(entry)

    @property
    def historian(self):
        """Get the Historian that this result set refers to"""
        return self._historian

    @property
    def archive_collection(self) -> archives.Collection:
        """Access the archive these results are from"""
        return self._archive_collection

    @property
    def query(self) -> expr.Query:
        """Get the query expression from this result set"""
        return self._query

    def distinct(self, key: Union[str, fields.Field]) -> Iterator:
        if isinstance(key, fields.Field):
            key = key.get_path()
        yield from self._archive_collection.distinct(key, filter=self._query.get_filter())

    def any(self) -> Optional[T]:
        """
        Return a single object from the result set.

        :return: an arbitrary object or `None` if there aren't any
        """
        query = self._query.copy()
        query.limit = 1
        query.sort = None

        results = tuple(self._archive_collection.find(**query.__dict__, **self._kwargs))
        return self._entry_factory(results[0])

    def one(self) -> Optional[T]:
        """Return one item from a result set containing at most one item.

        :raises NotOneError: Raised if the result set contains more than one item.
        :return: The object or `None` if there aren't any
        """
        # limit could be 1 due to slicing, for instance.
        query = self._query.copy()
        if query.limit is not None and query.limit > 2:
            query.limit = 2
        results = tuple(self._archive_collection.find(**query.__dict__, **self._kwargs))
        if not results:
            return None

        if len(results) > 1:
            raise exceptions.NotOneError('one() used with more than one result available')

        return self._entry_factory(results[0])

    def count(self) -> int:
        """Return the number of entries in the result set"""
        return self._archive_collection.count(self._query.get_filter(), **self._kwargs)

    def _project(self, *field: str) -> Iterator:
        """Get raw fields from the record dictionary"""
        projection = {name: 1 for name in field}
        for entry in self._archive_collection.find(**self._query.__dict__,
                                                   projection=projection,
                                                   **self._kwargs):
            yield entry


class EntriesCollection(Generic[T]):
    """A collection of archive entries.  This is the base class but it can be specialised to provide specific
    functionality for a given collection"""

    def __init__(self, historian, archive_collection: archives.Collection,
                 entry_factory: Callable[[dict], T]):
        self._historian = historian
        self._archive_collection = archive_collection
        self._entry_factory = entry_factory
        self._result_set_factory = ResultSet

    def find(
            self,  # pylint: disable=redefined-builtin
            *filter: expr.FilterSpec,
            obj_type=None,
            obj_id=None,
            version: int = -1,
            state=None,
            meta: dict = None,
            extras: dict = None,
            sort=None,
            limit=None,
            skip=0) -> ResultSet[T]:
        """Query the collection returning a result set"""
        query = self._prepare_query(*filter,
                                    obj_type=obj_type,
                                    obj_id=obj_id,
                                    version=version,
                                    state=state,
                                    extras=extras)
        query.sort = sort
        query.limit = limit
        query.skip = skip
        return self._result_set_factory(self._historian,
                                        self._archive_collection,
                                        query,
                                        kwargs=dict(meta=meta),
                                        entry_factory=self._entry_factory)

    def distinct(  # pylint: disable=redefined-builtin
            self,
            key: str,
            *filter: expr.FilterSpec,
            obj_type=None,
            obj_id=None,
            version: int = -1,
            state=None,
            extras: dict = None) -> Iterator:
        """Get the distinct values for the given key, optionally restricting to a subset of results
        """
        yield from self.find(*filter,
                             obj_type=obj_type,
                             obj_id=obj_id,
                             version=version,
                             state=state,
                             extras=extras).distinct(key)

    def get(self, entry_id) -> T:
        return self._entry_factory(self._archive_collection.get(entry_id))

    def _prepare_query(self,
                       *expression,
                       obj_type=None,
                       obj_id=None,
                       version: int = -1,
                       state=None,
                       extras: dict = None) -> expr.Query:
        """Prepare a query filter expression from the passed filter criteria"""
        query = expr.Query(*expression)

        if obj_type is not None:
            query.append(self._get_type_expr(obj_type))

        if obj_id is not None:
            query.append(self._get_obj_id_expr(obj_id))

        if version is not None and version != -1:
            query.append(records.DataRecord.version == version)

        if state is not None:
            if isinstance(state, dict):
                result = flatten_filter('state', state)
                query.extend(result)
            else:
                query.append(records.DataRecord.state == state)

        if extras is not None:
            if not isinstance(extras, dict):
                raise TypeError("extras must be a dict, got '{}'".format(extras))

            result = flatten_filter('extras', extras)
            query.extend(result)

        return query

    def _get_obj_id_expr(self, obj_id) -> expr.Expr:
        if isinstance(obj_id, expr.Expr):
            oper = obj_id
        else:
            if isinstance(obj_id, list):
                oper = expr.In(list(map(self._historian._prepare_obj_id, obj_id)))  # pylint: disable=protected-access
            else:
                oper = expr.Eq(self._historian._prepare_obj_id(obj_id))  # pylint: disable=protected-access

        return expr.Comparison(records.DataRecord.obj_id, oper)

    def _get_type_expr(self, obj_type) -> expr.Expr:
        if isinstance(obj_type, expr.Expr):
            oper = obj_type
        else:
            if isinstance(obj_type, list):
                oper = expr.In(list(map(self._historian._prepare_type_id, obj_type)))  # pylint: disable=protected-access
            else:
                oper = expr.Eq(self._historian._prepare_type_id(obj_type))  # pylint: disable=protected-access

        return expr.Comparison(records.DataRecord.type_id, oper)


class ObjectCollection(EntriesCollection[object]):
    """A collection of objects"""

    def __init__(
        self,
        historian,
        archive_collection: archives.Collection,
        record_factory: Callable[[dict], records.DataRecord],
        obj_loader: Callable[[records.DataRecord], object],
    ):
        super().__init__(
            historian,
            archive_collection,
            self._create_object,
        )
        self._record_factory = record_factory
        self._obj_loader = obj_loader
        self._records = EntriesCollection(self._historian, archive_collection, record_factory)

    def get(self, entry_id) -> object:
        return self._create_object(self._archive_collection.get(entry_id))

    @property
    def records(self) -> EntriesCollection[records.DataRecord]:
        """Access the records directly"""
        return self._records

    def _create_object(self, archive_record: dict):
        # Translate to a data record and then create the object
        return self._obj_loader(self._record_factory(archive_record))


def flatten_filter(entry_name: str, query) -> list:
    """Expand nested search criteria, e.g. state={'color': 'red'} -> {'state.colour': 'red'}"""
    if isinstance(query, dict):
        transformed = _transform_query_keys(query, entry_name)
        flattened = [{key: value} for key, value in transformed.items()]
    else:
        flattened = [{entry_name: query}]

    return flattened


@functools.singledispatch
def _transform_query_keys(entry, prefix: str = ''):  # pylint: disable=unused-argument
    """Transform a query entry into the correct syntax given a global prefix and the entry itself"""
    return entry


@_transform_query_keys.register(list)
def _(entry: list, prefix: str = ''):
    return [_transform_query_keys(value, prefix) for value in entry]


@_transform_query_keys.register(dict)
def _(entry: dict, prefix: str = ''):
    transformed = {}
    for key, value in entry.items():
        if key.startswith('$'):
            if key in ('$and', '$not', '$nor', '$or'):
                transformed[key] = _transform_query_keys(value, prefix)
            else:
                update = {prefix: {key: value}} if prefix else {key: value}
                transformed.update(update)
        else:
            to_join = [prefix, key] if prefix else [key]
            # Don't pass the prefix on, we've consumed it here
            transformed['.'.join(to_join)] = _transform_query_keys(value)

    return transformed



"""Module that contains aggregation operations"""


def and_(*conditions) -> dict:
    """Helper that produces query dict for AND of multiple conditions"""
    if len(conditions) == 1:
        return conditions[0]

    return {'$and': list(conditions)}


def eq_(one, other) -> dict:
    """Helper that produces mongo aggregation dict for two items being equal"""
    return {'$eq': [one, other]}


def in_(*possibilities) -> dict:
    """Tests if a value is one of the possibilities"""
    return {'$in': possibilities}



import pytest

from mincepy import expr

# pylint: disable=pointless-statement, invalid-name


def test_expr_types_and_filters():
    """Test the query filters for expressions"""
    name_eq = expr.Comparison('name', expr.Eq('frank'))
    age_gt = expr.Comparison('age', expr.Gt(38))

    # Check all operators
    for expression in expr.Operator.__subclasses__():
        assert expression.oper.startswith('$')
        assert expression(123).dict() == {expression.oper: 123}

    # Check all logicals
    for list_expr in expr.Logical.__subclasses__():
        assert list_expr.oper.startswith('$')
        if issubclass(list_expr, expr.WithListOperand):
            assert list_expr([name_eq, age_gt]).dict() == \
                   {list_expr.oper: [name_eq.dict(), age_gt.dict()]}

            # Check that the the passthrough for list expressions work
            assert list_expr([name_eq]).dict() == name_eq.dict()

            with pytest.raises(TypeError):
                list_expr('non list')
            with pytest.raises(TypeError):
                list_expr(['non expression'])
        else:
            assert list_expr(name_eq).dict() == {list_expr.oper: name_eq.dict()}
            with pytest.raises(TypeError):
                list_expr(['non expression'])

    assert expr.Empty().dict() == {}

    with pytest.raises(TypeError):
        # Comparison takes an operator, not a string
        expr.Comparison('my_field', 'oper')


def test_expr_and_or():
    """Test expression and/or methods"""
    name_eq = expr.Comparison('name', expr.Eq('frank'))
    age_gt = expr.Comparison('age', expr.Gt(38))

    anded = name_eq & age_gt
    assert anded.dict() == {'$and': [name_eq.dict(), age_gt.dict()]}

    ored = name_eq | age_gt
    assert ored.dict() == {'$or': [name_eq.dict(), age_gt.dict()]}

    with pytest.raises(TypeError):
        name_eq & 'hello'

    with pytest.raises(TypeError):
        name_eq | 'goodbye'

    # Test fusing
    assert (anded & anded).operand == [name_eq, age_gt, name_eq, age_gt]
    assert (anded | anded).operand != [name_eq, age_gt, name_eq, age_gt]
    assert (ored | ored).operand == [name_eq, age_gt, name_eq, age_gt]
    assert (ored & ored).operand != [name_eq, age_gt, name_eq, age_gt]


def test_build_expr():
    """Test building an expression from a query dictionary"""
    name_eq = expr.Comparison('name', expr.Eq('frank'))
    assert expr.build_expr(name_eq) is name_eq

    assert isinstance(expr.build_expr({}), expr.Empty)

    assert isinstance(expr.build_expr({'name': 'tom', 'age': 54}), expr.And)


def test_query_overlapping_filter_keys():
    gt_24 = expr.Comparison('age', expr.Gt(24))
    lt_38 = expr.Comparison('age', expr.Lt(38))
    compound1 = gt_24 & lt_38
    compound2 = gt_24 & lt_38
    query_filter = expr.Query(compound1, compound2).get_filter()
    assert query_filter == {'$and': [expr.query_expr(compound1), expr.query_expr(compound2)]}



import uuid

import pytest

import mincepy
from mincepy import testing

# pylint: disable=too-few-public-methods


def test_type_helper(historian: mincepy.Historian):
    """Check that a type helper can be used to make a non-historian compatible type compatible"""

    class Bird:

        def __init__(self, specie='hoopoe'):
            self.specie = specie

    class BirdHelper(mincepy.TypeHelper):
        TYPE = Bird
        TYPE_ID = uuid.UUID('5cc59e03-ea5d-43ff-8814-3b6f2e22cd76')

        specie = mincepy.field()

    bird = Bird()
    with pytest.raises(TypeError):
        historian.save(bird)

    # Now register the helper...
    historian.register_type(BirdHelper())
    # ...and we should be able to save
    assert historian.save(bird) is not None


def test_transaction_snapshots(historian: mincepy.Historian):

    class ThirdPartyPerson:
        """A class from a third party library"""

        def __init__(self, name):
            self.name = name

    class PersonHelper(mincepy.TypeHelper):
        TYPE_ID = uuid.UUID('62d8c767-14bc-4437-a9a3-ca5d0ce65d9b')
        TYPE = ThirdPartyPerson
        INJECT_CREATION_TRACKING = True

        def yield_hashables(self, obj, hasher):
            yield from hasher.yield_hashables(obj.name)

        def eq(self, one, other) -> bool:
            return one.name == other.name

        def save_instance_state(self, obj, saver):
            return obj.name

        def load_instance_state(self, obj, saved_state, loader):
            obj.name = saved_state.name

    person_helper = PersonHelper()
    historian.register_type(person_helper)

    person_maker = mincepy.Process('person maker')

    with person_maker.running():
        martin = ThirdPartyPerson('Martin')

    historian.save(martin)
    assert historian.created_by(martin) == historian.get_obj_id(person_maker)


class Boat:

    def __init__(self, make: str, length: float, owner: testing.Person = None):
        self.make = make
        self.length = length
        self.owner = owner


class BoatHelper(mincepy.TypeHelper):
    TYPE_ID = uuid.UUID('4d82b67a-dbcb-4388-b20e-8542c70491d1')
    TYPE = Boat

    # Describe how to store the properties
    make = mincepy.field()
    length = mincepy.field()
    owner = mincepy.field(ref=True)


def test_simple_helper(historian: mincepy.Historian):
    historian.register_type(BoatHelper())

    jenneau = Boat('jenneau', 38.9)
    jenneau_id = historian.save(jenneau)
    del jenneau

    jenneau = historian.load(jenneau_id)
    assert jenneau.make == 'jenneau'
    assert jenneau.length == 38.9

    # Now check that references work
    martin = testing.Person('martin', 35)
    jenneau.owner = martin
    historian.save(jenneau)
    del jenneau

    jenneau = historian.load(jenneau_id)
    assert jenneau.owner is martin


class Powerboat(Boat):
    TYPE_ID = uuid.UUID('924ef5b2-ce20-40b0-8c98-4da470f6c2c3')
    horsepower = mincepy.field()

    def __init__(self, make: str, length: float, horsepower: float, owner: testing.Person = None):
        super().__init__(make, length, owner)
        self.horsepower = horsepower


class PowerboatHelper(BoatHelper):
    TYPE_ID = uuid.UUID('924ef5b2-ce20-40b0-8c98-4da470f6c2c3')
    TYPE = Powerboat

    horsepower = mincepy.field()


def test_subclass_helper(historian: mincepy.Historian):
    historian.register_type(PowerboatHelper())

    quicksilver = Powerboat('quicksilver', length=7.0, horsepower=115)
    quicksilver_id = historian.save(quicksilver)
    del quicksilver

    quicksilver = historian.load(quicksilver_id)
    assert quicksilver.make == 'quicksilver'
    assert quicksilver.length == 7.
    assert quicksilver.horsepower == 115

    martin = testing.Person('martin', 35)
    quicksilver.owner = martin
    historian.save(quicksilver)
    del quicksilver

    quicksilver = historian.load(quicksilver_id)
    assert quicksilver.owner is martin
